import ast
import os
import shutil
from pathlib import Path

import tensorflow as tf
from tensorflow.python.platform import gfile
from tensorflow.python.tools import freeze_graph
from miso.deploy.model_info import ModelInfo
import tensorflow.keras.backend as K
import tempfile
import lxml.etree as ET
import numpy as np
import tf2onnx


def remove(path):
    """ param <path> could either be relative or absolute. """
    if os.path.isfile(path):
        os.remove(path)  # remove the file
    elif os.path.isdir(path):
        shutil.rmtree(path)  # remove dir and all contains
    # else:
    #     raise ValueError("file {} is not a file or dir.".format(path))


def convert_to_inference_mode(model, model_factory):
    # Trick to get around bugs in tensorflow 1.14.0
    # https://github.com/tensorflow/tensorflow/issues/31331#issuecomment-518655879
    with tempfile.TemporaryDirectory() as dirpath:
        weights_filename = os.path.join(dirpath, "weights.h5")
        model.save_weights(weights_filename)
        K.clear_session()
        K.set_learning_phase(0)
        model = model_factory()
        model.load_weights(weights_filename)
        remove(weights_filename)
    return model


def convert_to_inference_mode_tf2(model, model_factory):
    # Trick to get around bugs in tensorflow 1.14.0
    # https://github.com/tensorflow/tensorflow/issues/31331#issuecomment-518655879
    with tempfile.TemporaryDirectory() as dirpath:
        weights_filename = os.path.join(dirpath, "weights.tf")
        model.save_weights(weights_filename)
        K.clear_session()
        K.set_learning_phase(0)
        model = model_factory()
        model.load_weights(weights_filename)
        remove(weights_filename)
    return model


def save(model, save_dir):
    tf.saved_model.simple_save(K.get_session(),
                               save_dir,
                               inputs={"input": model.inputs[0]},
                               outputs={"output": model.outputs[0]})


def freeze(model, save_dir):
    # if metadata is not None:
    #     metadata_tensor = K.constant(metadata.to_xml(), name="metadata", dtype='string')
    #     model = Model(model.inputs[0], [model.outputs[0], metadata_tensor])
    # K.set_learning_phase(0)

    # Save using simple save
    save(model, save_dir)
    freeze_graph.freeze_graph(None,
                              None,
                              None,
                              None,
                              model.outputs[0].op.name,
                              None,
                              None,
                              os.path.join(save_dir, "frozen_model.pb"),
                              False,
                              "",
                              input_saved_model_dir=save_dir)
    # Delete saved models
    remove(os.path.join(save_dir, "saved_model.pb"))
    remove(os.path.join(save_dir, "variables"))

def load_frozen_model(source: str,
                      input_tensor="input_1:0",
                      output_tensor="conv2d_23/Sigmoid:0",
                      session=None):
    if not os.path.exists(source):
        raise FileNotFoundError('The graph file was not found on the system.\nThe path was: ' + source)
    # Load network
    if session is None:
        session = K.get_session()
    with gfile.Open(source, 'rb') as f:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(f.read())
        session.graph.as_default()
        tf.import_graph_def(graph_def, name='')
    # Information
    # names = [n.name for n in tf.get_default_graph().as_graph_def().node]
    # for n in names:
    #     print(n)
    # Input / output tensors
    input = session.graph.get_tensor_by_name(input_tensor)
    output = session.graph.get_tensor_by_name(output_tensor)
    return session, input, output


def load_from_xml(filename, session=None):
    project = ET.parse(filename).getroot()

    protobuf = project.find('protobuf').text
    print(f"Loading model from {filename}")
    print("- protobuf: " + protobuf)

    input = None
    output = None
    img_size = np.zeros(3, dtype=int)
    cls_labels = []

    list_xml = project.find('labels')
    print("- labels:")
    for i, entry_xml in enumerate(list_xml.iter('label')):
        code = entry_xml.find('code').text
        cls_labels.append(code)
        print(f"  - {code}")

    list_xml = project.find('inputs')
    for i, entry_xml in enumerate(list_xml.iter('input')):
        if i == 0:
            input_name = entry_xml.find('operation').text + ":0"
            img_size[0] = int(entry_xml.find('height').text)
            img_size[1] = int(entry_xml.find('width').text)
            img_size[2] = int(entry_xml.find('channels').text)

    list_xml = project.find('outputs')
    for i, entry_xml in enumerate(list_xml.iter('output')):
        if i == 0:
            output_name = entry_xml.find('operation').text + ":0"

    # Find the 'cnn' tag within 'params'
    cnn_tag = project.find('.//params/cnn')

    # Function to safely evaluate the OrderedDict string
    def eval_ordered_dict(od_str):
        try:
            return ast.literal_eval(od_str)
        except ValueError as e:
            print(f"Error evaluating OrderedDict: {e}")
            return None

    # Evaluate and store the OrderedDict from the 'cnn' tag, if it exists
    cnn_params = None
    if cnn_tag is not None and cnn_tag.text:
        cnn_params = eval_ordered_dict(cnn_tag.text.strip())

    full_protobuf_path = os.path.join(os.path.dirname(filename), protobuf)

    print(f"- input: {input_name}")
    print(f"  - height: {img_size[0]}")
    print(f"  - width: {img_size[1]}")
    print(f"  - channels: {img_size[2]}")
    print(f"- output: {output_name}")

    model = load_frozen_model_tf2(full_protobuf_path, input_name, output_name)
    return model, img_size, cnn_params["img_type"], cls_labels


import onnxruntime as rt

def load_onnx_from_xml(filename):
    project = ET.parse(filename).getroot()

    protobuf = str(list(Path(filename).parent.glob("*.onnx"))[0])
    print(f"Loading model from {filename}")
    print(f"- protobuf: {protobuf}")



    input = None
    output = None
    img_size = np.zeros(3, dtype=int)
    cls_labels = []

    list_xml = project.find('labels')
    print("- labels:")
    for i, entry_xml in enumerate(list_xml.iter('label')):
        code = entry_xml.find('code').text
        cls_labels.append(code)
        print(f"  - {code}")

    list_xml = project.find('inputs')
    for i, entry_xml in enumerate(list_xml.iter('input')):
        if i == 0:
            input_name = entry_xml.find('operation').text + ":0"
            img_size[0] = int(entry_xml.find('height').text)
            img_size[1] = int(entry_xml.find('width').text)
            img_size[2] = int(entry_xml.find('channels').text)

    list_xml = project.find('outputs')
    for i, entry_xml in enumerate(list_xml.iter('output')):
        if i == 0:
            output_name = entry_xml.find('operation').text + ":0"

    # Find the 'cnn' tag within 'params'
    cnn_tag = project.find('.//params/cnn')

    # Function to safely evaluate the OrderedDict string
    def eval_ordered_dict(od_str):
        try:
            return ast.literal_eval(od_str)
        except ValueError as e:
            print(f"Error evaluating OrderedDict: {e}")
            return None

    # Evaluate and store the OrderedDict from the 'cnn' tag, if it exists
    cnn_params = None
    if cnn_tag is not None and cnn_tag.text:
        cnn_params = eval_ordered_dict(cnn_tag.text.strip())

    full_protobuf_path = os.path.join(os.path.dirname(filename), protobuf)

    print(f"- input: {input_name}")
    print(f"  - height: {img_size[0]}")
    print(f"  - width: {img_size[1]}")
    print(f"  - channels: {img_size[2]}")
    print(f"- output: {output_name}")


    sess = rt.InferenceSession(str(protobuf))

    return sess, img_size, cnn_params["img_type"], cls_labels




"""
Adapted from https://leimao.github.io/blog/Save-Load-Inference-From-TF2-Frozen-Graph/
"""
from tensorflow.python.framework.convert_to_constants import convert_variables_to_constants_v2


def wrap_frozen_graph_tf2(graph_def, inputs, outputs, print_graph=False):
    def _imports_graph_def():
        tf.compat.v1.import_graph_def(graph_def, name="")

    wrapped_import = tf.compat.v1.wrap_function(_imports_graph_def, [])
    import_graph = wrapped_import.graph
    if print_graph == True:
        print("-" * 50)
        print("Frozen model layers: ")
        layers = [op.name for op in import_graph.get_operations()]
        for layer in layers:
            print(layer)
        print("-" * 50)
    return wrapped_import.prune(
        tf.nest.map_structure(import_graph.as_graph_element, inputs),
        tf.nest.map_structure(import_graph.as_graph_element, outputs))


def save_frozen_model_tf2(model, out_dir, filename):
    print("-" * 80)
    print("Freezing model")
    # Convert Keras model to ConcreteFunction
    full_model = tf.function(lambda x: model(x))
    full_model = full_model.get_concrete_function(
        tf.TensorSpec(model.inputs[0].shape, model.inputs[0].dtype))

    # Get frozen ConcreteFunction
    frozen_func = convert_variables_to_constants_v2(full_model)
    frozen_func.graph.as_graph_def()

    # layers = [op.name for op in frozen_func.graph.get_operations()]
    # for layer in layers:
    #     print(layer)

    print("- model inputs: {}".format(frozen_func.inputs))
    print("- model outputs: {}".format(frozen_func.outputs))

    # Save frozen graph from frozen ConcreteFunction to hard drive
    # logdir="./frozen_models",
    print(out_dir)
    print(filename)
    tf.io.write_graph(graph_or_graph_def=frozen_func.graph,
                      logdir=os.path.join(out_dir),
                      name=filename,
                      as_text=False)
    return frozen_func


def save_model_as_onnx(model, input_name, input_shape, out_dir, opset=13):
    spec = (tf.TensorSpec(input_shape, tf.float32, name=input_name),)
    output_path = os.path.join(out_dir, "model.onnx")
    model_proto, _ = tf2onnx.convert.from_keras(model, input_signature=spec, opset=opset, output_path=output_path)
    output_names = [n.name for n in model_proto.graph.output]
    print(output_names)


def load_frozen_model_tf2(filepath, inputs, outputs):
    with tf.io.gfile.GFile(filepath, "rb") as f:
        graph_def = tf.compat.v1.GraphDef()
        loaded = graph_def.ParseFromString(f.read())
    # Wrap frozen graph to ConcreteFunctions
    frozen_func = wrap_frozen_graph_tf2(graph_def=graph_def,
                                        inputs=inputs,
                                        outputs=outputs,
                                        print_graph=False)
    # print("-" * 80)
    # print("Frozen model inputs: ")
    # print(frozen_func.inputs)
    # print("Frozen model outputs: ")
    # print(frozen_func.outputs)
    return frozen_func



import tensorflow as tf
import tensorflow.keras.backend as K
from tensorflow.keras.preprocessing.image import ImageDataGenerator


def image_generator(images,
                    onehots,
                    batch_size):
    """
    Create a simple generator without modifying the images

    :param images: Images to generate
    :param onehots: Onehot encoding of target class
    :param batch_size: Batch size for training
    :return:
    """
    datagen = ImageDataGenerator()
    return keras_augmented_image_generator(images, onehots, batch_size, datagen)


def keras_augmented_image_generator(images,
                                    onehots,
                                    batch_size,
                                    datagen):
    """
    Create a generator from with augmentation via an ImageDataGenerator
    The generator can then be used for training in model.fit_generator

    The ImageDataGenerator should be configured with the desired augmentations.
    Note: when brightness is used, rescale=1./255 must also be used
    (assuming a desired output range of [0,1] as currently brightness augmentation internally rescales to [0,255]

    :param images: Images to augment
    :param onehots: Onehot encoding of target class
    :param batch_size: Batch size for training
    :param datagen: The ImageDataGenerator with augmentation options configured
    :return:
    """
    return datagen.flow(images,
                        onehots,
                        batch_size=batch_size,
                        shuffle=True)


def tf_augmented_image_generator(images,
                                 onehots,
                                 batch_size,
                                 map_fn,
                                 shuffle_size=1000,
                                 num_parallel_calls=tf.data.experimental.AUTOTUNE):
    """
    Create a generator suing a tf.data.Dataframe with augmentation via a map function.
    The generator can then be used for training in model.fit_generator

    The map function must consist of tensorflow operators (not numpy).

    On Windows machines this will lead to faster augmentation, as there are some
    problems performing augmentation in parallel when multiprocessing is enabled in
    in model.fit / model.fit_generator and the default Keras numpy-based augmentated is used,
    e.g. in ImageDataGenerator

    :param images: Images to augment
    :param onehots: Onehot encoding of target class
    :param batch_size: Batch size for training
    :param map_fn: The augmentation map function
    :param shuffle_size: Batch size of images shuffled. Smaller values reduce memory consumption.
    :param num_parallel_calls: Number of calls in parallel, default is automatic tuning.
    :return:
    """
    # Get shapes from input data
    img_size = images.shape
    img_size = (None, img_size[1], img_size[2], img_size[3])
    onehot_size = onehots.shape
    onehot_size = (None, onehot_size[1])
    images_tensor = tf.placeholder(tf.float32, shape=img_size)
    onehots_tensor = tf.placeholder(tf.float32, shape=onehot_size)
    # Create dataset
    dataset = tf.data.Dataset.from_tensor_slices((images_tensor, onehots_tensor))
    if map_fn is not None:
        dataset = dataset.map(lambda x, y: (map_fn(x), y), num_parallel_calls=num_parallel_calls)
    dataset = dataset.shuffle(shuffle_size, reshuffle_each_iteration=True).repeat()
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(1)
    iterator = dataset.make_initializable_iterator()
    init_op = iterator.initializer
    next_val = iterator.get_next()
    with K.get_session().as_default() as sess:
        sess.run(init_op, feed_dict={images_tensor: images, onehots_tensor: onehots})
        while True:
            inputs, labels = sess.run(next_val)
            yield inputs, labels

def tf_vector_generator(vectors,
                         onehots,
                         batch_size,
                         shuffle_size=1000,
                         num_parallel_calls=tf.data.experimental.AUTOTUNE):
    # Get shapes from input data
    vec_size = vectors.shape
    vec_size = (None, vec_size[1])
    onehot_size = onehots.shape
    onehot_size = (None, onehot_size[1])
    vectors_tensor = tf.placeholder(tf.float32, shape=vec_size)
    onehots_tensor = tf.placeholder(tf.float32, shape=onehot_size)
    # Create dataset
    dataset = tf.data.Dataset.from_tensor_slices((vectors_tensor, onehots_tensor))
    dataset = dataset.shuffle(shuffle_size, reshuffle_each_iteration=True).repeat()
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(1)
    iterator = dataset.make_initializable_iterator()
    init_op = iterator.initializer
    next_val = iterator.get_next()
    with K.get_session().as_default() as sess:
        sess.run(init_op, feed_dict={vectors_tensor: vectors, onehots_tensor: onehots})
        while True:
            inputs, labels = sess.run(next_val)
            yield inputs, labels


def image_generator_from_dataframe(dataframe,
                                   img_size,
                                   batch_size,
                                   cls_labels,
                                   datagen,
                                   color_mode='rgb'):
    """
    Creates a generator that loads images from information in a pandas dataframe.

    The dataframe must have at least two columns:
    - "filename" with the absolute path to the file
    - "cls" with the class label of each image (text)

    Images will be preprocessed using an ImageDataGenerator, resized to a fixed shape and converted to grayscale if desired.

    :param dataframe: Pandas dataframe with the image information
    :param img_size: Shape of to resize the images to, e.g. (128, 128)
    :param batch_size: Size of the generator batch
    :param cls_labels: List containing each class label
    :param datagen: The ImageDataGenerator for preprocessing
    :param color_mode: 'rgb' or 'grayscale' to produce 3 or 1 channel images, respectively
    :return:
    """
    return datagen.flow_from_dataframe(
        dataframe,
        x_col="filenames",
        y_col="cls",
        classes=cls_labels,
        target_size=img_size,
        batch_size=batch_size,
        color_mode=color_mode,
        interpolation='bilinear',
        class_mode='categorical')


def tf_augmented_image_generator_for_segmentation(images,
                                                  targets,
                                                  batch_size,
                                                  map_fn,
                                                  augment_targets=False,
                                                  shuffle_size=1000,
                                                  num_parallel_calls=tf.data.experimental.AUTOTUNE):
    # Get shapes from input data
    img_size = images.shape
    img_size = (None,) + img_size[1:]
    target_size = targets.shape
    target_size = (None,) + target_size[1:]
    images_tensor = tf.compat.v1.placeholder(tf.float32, shape=img_size)
    targets_tensor = tf.compat.v1.placeholder(tf.float32, shape=target_size)

    # Create dataset
    dataset = tf.data.Dataset.from_tensor_slices((images_tensor, targets_tensor))
    if map_fn is not None:
        if augment_targets is False:
            dataset = dataset.map(lambda x, y: (map_fn(x), y), num_parallel_calls=num_parallel_calls)
        else:
            dataset = dataset.map(lambda x, y: map_fn(x, y), num_parallel_calls=num_parallel_calls)
    dataset = dataset.shuffle(shuffle_size, reshuffle_each_iteration=True).repeat()
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(1)

    iterator = dataset.make_initializable_iterator()
    init_op = iterator.initializer
    next_val = iterator.get_next()

    with K.get_session().as_default() as sess:
        sess.run(init_op, feed_dict={images_tensor: images, targets_tensor: targets})
        while True:
            inputs, labels = sess.run(next_val)
            yield inputs, labels



from pathlib import Path

import matplotlib.pyplot as plt
from sklearn.neighbors import KNeighborsClassifier
import numpy as np
from sklearn.utils.extmath import weighted_mode
from sklearn.preprocessing import normalize
import os

from tqdm import tqdm


def find_and_save_mislabelled(images,
                              vectors,
                              cls,
                              cls_labels,
                              image_names,
                              output_dir,
                              num_neighbours=11):
    """
    Uses nearest neighbour to classify the vectors and checks this matches the actual classifications.
    If not, prints the most similar other images. Also writes to a csv file with this list.
    Args:
        images: Images (for use in plots)
        vectors: Vectors for each image
        cls: Class label for each image
        cls_labels: List of class cls in order [class_0, class_1, ..., class_N]
        image_names: List of ids for each image, e.g. their filenames
        output_dir: Directory to save the results
        num_neighbours: Number of neighbours to use for the kNN classification
    Returns: None
    """
    # Normalise vectors
    vectors = normalize(vectors, axis=1)
    # Nearest neighbours fit
    neigh = KNeighborsClassifier(n_neighbors=num_neighbours + 1, algorithm='brute')
    neigh.fit(vectors, cls)

    # Find the k nearest vectors
    # The first will be the same vector as passed in so ignore it
    knn = neigh.kneighbors(vectors, n_neighbors=num_neighbours + 1)
    distance = knn[0][:, 1:]
    idx = knn[1][:, 1:]
    # Get the predicted class from knn (ignoring the same vector)
    knn_cls = cls[idx]
    wm = weighted_mode(knn_cls, np.max(distance) - distance, axis=1)
    pred_cls = wm[0].astype(int).flatten()
    pred_w = wm[1]

    # Get the index of images with a different k-NN prediction to their label
    diff_idx = np.where(cls != pred_cls)[0]

    # Write a CSV of the mislabelled
    with open(os.path.join(output_dir, "mislabeled.csv"), 'w+') as f:
        f.write("filename, label, predicted_label\n")
        for i in diff_idx:
            f.write("{},{},{}\n".format(image_names[i], cls_labels[cls[i]], cls_labels[pred_cls[i]]))

    if num_neighbours == 1:
        nx = 2
        ny = 1
    elif num_neighbours == 2:
        nx = 3
        ny = 1
    elif num_neighbours == 3:
        nx = 2
        ny = 2
    elif num_neighbours == 5:
        nx = 3
        ny = 2
    else:
        nx = 4
        ny = int(np.ceil((num_neighbours + 1) / 4))

    # Plot each image
    for im_idx in tqdm(diff_idx):
        fig, axes = plt.subplots(nrows=ny, ncols=nx, figsize=(2 * nx, 2 * ny + 1),
                                 gridspec_kw={"top": (2 * ny + 0.5) / (2 * ny + 1),
                                              "left": 0.05,
                                              "right": 0.95,
                                              "bottom": 0.5 / (2 * ny + 1)})
        plt.set_cmap('gray')
        fig.set_facecolor("white")
        for i, ax in enumerate(axes.flat):
            if i == 0:
                image = images[im_idx].astype(np.uint8)
                if images.shape[3] == 1:
                    ax.imshow(image[:, :, 0])
                else:
                    ax.imshow(image)
                if cls_labels is None:
                    xlabel = "{0}\n({1})?".format(cls[im_idx], pred_cls[im_idx])
                else:
                    xlabel = "{0}\n{1}?".format(cls_labels[cls[im_idx]], cls_labels[pred_cls[im_idx]])
                ax.set_xlabel(xlabel)
            else:
                kidx = idx[im_idx, i - 1]
                image = images[kidx].astype(np.uint8)
                if images.shape[3] == 1:
                    ax.imshow(image[:, :, 0])
                else:
                    ax.imshow(image)
                if cls_labels is None:
                    xlabel = "{0}".format(cls[kidx])
                else:
                    xlabel = "{0}".format(cls_labels[cls[kidx]])
                ax.set_xlabel(xlabel)
            # Remove ticks from the plot.
            ax.set_xticks([])
            ax.set_yticks([])
        # Tight layout looks nicer
        fig.tight_layout()
        plt.suptitle(image_names[im_idx])
        # print("- actual: {} predicted: {} - {}".format(cls_labels[cls[im_idx]], cls_labels[pred_cls[im_idx]], image_names[im_idx]))

        filename = str(Path(image_names[im_idx]).stem) + ".png"
        path = Path(output_dir) / "mislabeled" / cls_labels[cls[im_idx]]
        path.mkdir(parents=True, exist_ok=True)
        plt.savefig(path / filename)
        plt.clf()
        plt.close('all')



from typing import NamedTuple

import numpy as np
from glob import glob
import os
from collections import OrderedDict
import platform
import pandas as pd
import xml.etree.ElementTree as ET

from miso.data.download import download_images

from pathlib import Path



def ls(source_dir):
    return sorted(glob(os.path.join(source_dir, "*")))


def parse_directory(source_dir, skip='~', has_classes=True):
    """
    Parses a directory consisting of subdirectories named by class and stores the image filenames in a dictionary
    :param source_dir: The directory to parse
    :param verbose: Print out the progress
    :param skip: If a subdirectory starts with this character it will be skipped ('_' and '.' are skipped always)
    :param has_classes: The files are arranged in subdirectories by class name
    :return: Dictionary with class names for the keys and lists of filenames for the values
    """
    if has_classes:
        sub_dirs = sorted(glob(os.path.join(source_dir, "*")))
    else:
        sub_dirs = [source_dir]
    filenames = OrderedDict()
    for sub_dir in sub_dirs:
        if os.path.isdir(sub_dir) is False:
            continue
        # Get the directory name
        sub_name = os.path.basename(sub_dir)
        # Skip directories starting with ~
        if sub_name.startswith(skip) or \
            sub_name.startswith('_') or \
            sub_name.startswith('.'):
            continue
        files = []
        for ext in ["*.jpg", "*.jpeg", "*.png", "*.tif", "*.tiff", "*.bmp"]:
            sub_files1 = sorted(glob(os.path.join(sub_dir, ext)))
            files.extend(sub_files1)
            if platform.system() == 'Linux':
                sub_files2 = sorted(glob(os.path.join(sub_dir, ext.upper())))
                files.extend(sub_files2)
        # Add to dictionary
        if has_classes:
            filenames[sub_name] = files
        else:
            filenames['null'] = files
    return filenames


def parse_csv(csv_file, source_dir, file_idx=0, cls_idx=1, cls_label_idx=2, verbose=True):
    if verbose:
        print("Parsing csv file {} for directory {}".format(csv_file, source_dir))
    df = pd.read_csv(csv_file)
    num_classes = np.max(df.iloc[:, cls_idx]) + 1
    cls_labels = [df.loc[df.iloc[:, cls_idx] == i].iloc[0, cls_label_idx] for i in range(num_classes)]
    filenames = OrderedDict()
    for i in range(num_classes):
        if verbose:
            print("- {}".format(cls_labels[i]), end='')
        names = df.loc[df.iloc[:, cls_idx] == i].iloc[:, file_idx]
        paths = [os.path.join(source_dir, n) for n in names]
        filenames[cls_labels[i]] = paths
        if verbose:
            print(" ({} files)".format(len(paths)))
    return filenames


def parse_xml(xml_filename):
    project = ET.parse(xml_filename).getroot()

    filenames = []
    cls = []
    cls_labels = []

    filenames_dict = OrderedDict()

    images_xml = project.find('images')
    for i, image_xml in enumerate(images_xml.iter('image')):
        relfile = image_xml.find('source').find('filename').text
        if os.path.isabs(relfile):
            absfile = relfile
        else:
            absfile = os.path.abspath(os.path.join(os.path.dirname(xml_filename), relfile))
        if os.path.isfile(absfile) is False:
            continue
        filenames.append(absfile)

        cls_names = []
        cls_scores = []
        cls_base = image_xml.find('classifications')
        for cls_val in cls_base.iter('classification'):
            cls_names.append(cls_val.find('code').text)
            cls_scores.append(float(cls_val.find('value').text))
        cls.append(cls_names[np.argmax(cls_scores)])

    for taxon_xml in project.find('taxons').iter('taxon'):
        if taxon_xml.find('isClass').text == 'true':
            cls_labels.append(taxon_xml.find('code').text)
    cls_labels = sorted(cls_labels)

    df = pd.DataFrame({'filenames': filenames, 'cls': cls})
    for label in cls_labels:
        filenames_dict[label] = df.filenames[df.cls == label]
    return filenames_dict


class FilenamesDataset(object):
    def __init__(self,
                 source: str,
                 csv_idxs=(0, 1, 2),
                 has_classes=True):
        self.source = source
        self.csv_idxs = csv_idxs
        self.has_classes = has_classes
        self.cls_filenames = OrderedDict()
        self.filenames = None
        self.cls_labels = None
        self.cls = None
        self.cls_counts = None
        self.num_classes = None

    def load(self, min_count=0, map_others=False):
        print('-' * 80)
        print("Parsing source {}".format(self.source))
        print()

        # Load filenames
        if self.source.endswith(".xml"):
            self.cls_filenames = parse_xml(self.source)
        elif self.source.endswith(".csv"):
            self.cls_filenames = parse_csv(self.source, self.source, self.csv_idxs[0], self.csv_idxs[1], self.csv_idxs[2])
        elif self.source.startswith("http"):
            dir_for_download = os.path.join(str(Path.home()), 'miso_datasets')
            os.makedirs(dir_for_download, exist_ok=True)
            dir_path = download_images(self.source, dir_for_download)
            self.cls_filenames = parse_directory(dir_path, has_classes=self.has_classes)
        else:
            self.cls_filenames = parse_directory(self.source, has_classes=self.has_classes)

        # Check we found some images
        filenames = [v for key, val in self.cls_filenames.items() for v in val]
        if len(filenames) == 0:
            raise ValueError("! Did not find any images")

        # Display how many found
        print("Classes:")
        for k, v in self.cls_filenames.items():
            print("- {}: {} images".format(k, len(v)))
        print()

        # Apply min count
        if min_count > 0:
            print("Removing classes with less than {} images:".format(min_count))
            others = []
            filtered_cls_filenames = OrderedDict()
            was_min = False
            for k, v in self.cls_filenames.items():
                if len(v) < min_count:
                    others.extend(v)
                    print("- {}: {} images".format(k, len(v)))
                    was_min = True
                else:
                    filtered_cls_filenames[k] = v
            if was_min is False:
                print("- none found")
            # Put in others
            if map_others is True and len(others) > min_count:
                if 'others' in filtered_cls_filenames:
                    filtered_cls_filenames['others'].extend(others)
                else:
                    filtered_cls_filenames['others'] = others
                print("- these images have been places in the 'others' class, {} total".format(len(others)))
            self.cls_filenames = filtered_cls_filenames
            print()

        # Flatten filename and class dictionary to a list
        self.filenames = [v for key, val in self.cls_filenames.items() for v in val]
        self.cls_labels = list(self.cls_filenames.keys())
        self.cls = np.asarray([self.cls_labels.index(key) for key, val in self.cls_filenames.items() for v in val])
        self.cls_counts = np.asarray([len(val) for key, val in self.cls_filenames.items()])
        self.num_classes = len(self.cls_filenames.keys())




import pandas as pd
import os
import skimage.io as skio
from collections import OrderedDict
from glob import glob
from tqdm import tqdm
import argparse


def process_dir(input_dir, save_dir, campaign_name, species_filename=None, save_csv=True):
    # Find all lst files
    lst_filenames = sorted(glob(os.path.join(input_dir, "**", "*.lst"), recursive=True))
    print(lst_filenames)
    # Extract the base directories
    dirs = [os.path.dirname(fn) for fn in lst_filenames]
    # Only take unique ones
    dirs = sorted(list(set(dirs)))
    # dirs = [d for d in sorted(glob(os.path.join(input_dir, "*"))) if os.path.isdir(d)]
    df_master = None
    for d in dirs:
        # Directory
        if d == input_dir:
            run_name = os.path.basename(d).replace("/", "_").replace(" ", "_").replace("\\", "_")
        else:
            run_name = d[len(input_dir)+1:].replace("/", "_").replace(" ", "_").replace("\\", "_")
        df = process(d, save_dir, campaign_name, run_name, species_filename, save_csv=True)
        if df_master is None:
            df_master = df
            print("first")
            print(len(df_master))
        else:
            df_master = df_master.append(df)
            print("append")
            print(len(df_master))
    if save_csv:
        df_master.to_csv(os.path.join(save_dir, campaign_name, "{}_data.csv".format(campaign_name)), index=False)
    return df_master


def process(input_dir, save_dir, campaign_name, run_name, species_filename=None, save_csv=True, save_mask=False):
    # Load species conversion list
    if species_filename is not None and species_filename != "":
        # noms = pd.read_excel(species_filename, sheet_name=0, engine='openpyxl').dropna()
        doublons = pd.read_excel(species_filename, sheet_name=1, engine='openpyxl').dropna()
        # feuil1 = pd.read_excel(species_filename, sheet_name=2, engine='openpyxl').dropna()
    else:
        doublons = None

    # The directory where we will save the images
    im_save_dir = os.path.join(save_dir, campaign_name, "images", run_name)
    os.makedirs(im_save_dir, exist_ok=True)
    if save_mask:
        mask_save_dir = os.path.join(save_dir, campaign_name, "masks", run_name)
        os.makedirs(mask_save_dir, exist_ok=True)
    save_dir = os.path.join(save_dir, campaign_name, run_name)

    print('-' * 80)
    # print("Flowcam segmenter")
    # print('-' * 80)
    # print("- species XLSX: {}".format(species_filename))
    print("Campaign: {}".format(campaign_name))
    print("Sample: {}".format(run_name))
    print("- input directory: {}".format(input_dir))

    # Image data
    lst_filename = glob(os.path.join(input_dir, "*.lst"))
    if len(lst_filename) == 0:
        print("! No .lst file found in {}, skipping !".format(input_dir))
        return
    lst_filename = lst_filename[0]
    df = parse_image_list(lst_filename)

    # Classification data
    cla_filename = glob(os.path.join(input_dir, "*.cla"))
    if len(cla_filename) == 0 or os.path.getsize(cla_filename[0]) == 0:
        print("- classification (.cla) file is missing or empty, all images will be placed in \"unlabeled\" class")
        cls_dict = dict()
    else:
        cla_filename = cla_filename[0]
        cls_dict = parse_classifications(cla_filename)
        print("- classification filename: {}".format(cla_filename))
    print("- image data filename: {}".format(lst_filename))
    print("- output directory: {}".format(save_dir))
    print("Processing...")

    # Group the results by image
    df_grouped = df.groupby("collage_file")

    # Extra info to save
    df_filename = [""] * len(df)
    df_cls = [""] * len(df)
    df_campaign = [campaign_name] * len(df)
    df_sample = [run_name] * len(df)

    # Process each image
    for filename, group in tqdm(df_grouped):
        # Load the image
        im_filename = os.path.join(input_dir, filename)
        mask_filename = os.path.join(input_dir, filename[:-4] + "_bin.tif")

        if os.path.exists(im_filename):
            try:
                im = skio.imread(im_filename)
            except:
                print("Error opening image {}".format(im_filename))
                continue
        else:
            print("Image not found {}".format(im_filename))

        is_mask = False
        if save_mask:
            if os.path.exists(mask_filename):
                try:
                    mask = skio.imread(mask_filename)
                    is_mask = True
                except:
                    print("Error opening {}".format(mask_filename))
            else:
                print("Mask not found {}".format(mask_filename))

        # Cut each image out
        for row in group.iterrows():
            row_id = row[0]
            row = row[1]
            # Get image number
            id = row['id']
            # Get class
            if id in cls_dict:
                cls = cls_dict[id]
            else:
                cls = "unlabeled"
            # Modify class to correct if needed
            if doublons is not None and cls in doublons.iloc[:, 0].values:
                vals = doublons[doublons.iloc[:, 0] == cls]
                new_cls = vals.iloc[0, 1]
                cls = new_cls
            # Get image coordinates
            x = row['image_x']
            y = row['image_y']
            width = row['image_w']
            height = row['image_h']
            # Get the segmented image
            seg_im = im[y:y + height, x:x + width, ...]
            seg_im_filename = os.path.join(im_save_dir, cls, "{}_{}_{:08d}.png".format(campaign_name, run_name, id))
            os.makedirs(os.path.dirname(seg_im_filename), exist_ok=True)
            skio.imsave(seg_im_filename, seg_im)
            if is_mask:
                seg_mask = mask[y:y + height, x:x + width, ...]
                seg_mask_filename = os.path.join(mask_save_dir, cls, "{}_{}_{:08d}.png".format(campaign_name, run_name, id))
                os.makedirs(os.path.dirname(seg_mask_filename), exist_ok=True)
                skio.imsave(seg_mask_filename, seg_mask)
            df_cls[id-1] = cls
            df_filename[id-1] = seg_im_filename
    df.insert(0, 'filename', df_filename)
    df.insert(2, 'campaign', df_campaign)
    df.insert(3, 'sample', df_sample)
    df.insert(4, 'class', df_cls)
    if save_csv:
        df.to_csv(os.path.join(im_save_dir, "{}_{}_data.csv".format(campaign_name, run_name)), index=False)
    print("Complete!")
    return df


def parse_classifications(filename):
    cls_dict = OrderedDict()
    with open(filename, "r") as f:
        f.readline()
        f.readline()
        f.readline()
        num_classes = int(f.readline())
        for i in range(num_classes):
            cls_name = f.readline()[:-1]
            f.readline()
            f.readline()
            num_images = int(f.readline())
            for j in range(num_images):
                idx = int(f.readline())
                cls_dict[idx] = cls_name
    return cls_dict


def parse_image_list(filename):
    field_names = []
    with open(filename, "r") as f:
        f.readline()
        numfields = int(f.readline().split('|')[1])
        for i in range(numfields):
            field_names.append(f.readline().split('|')[0])
        # print(field_names)
    df = pd.read_csv(filename, '|', skiprows=numfields + 2, header=None)
    df.columns = field_names
    return df

#
# def process(folder):
#     cla_filename = glob(os.path.join(folder, "*.cla"))[0]
#     lst_filename = glob(os.path.join(folder, "*.lst"))[0]
#     run_id = os.path.basename(cla_filename)[:-4]
#
#     cls_dict = class_list(cla_filename)
#
#     field_names = []
#     with open(lst_filename, "r") as f:
#         numfields = int(f.readline().split('|')[1])
#         for i in range(numfields):
#             field_names.append(f.readline().split('|')[0])
#         print(field_names)
#     df = pd.read_csv(lst_filename, '|', skiprows=numfields + 2, header=None)
#     df.columns = field_names
#
#     last_image_name = None
#     last_image = None
#     for row in df.iterrows():
#         row_id = row[0]
#         row = row[1]
#         if row['collage_file'] != last_image_name:
#             last_image_name = row['collage_file']
#             last_image = skio.imread(os.path.join(folder, last_image_name))
#         id = row['id']
#         if id in cls_dict:
#             cls_name = cls_dict[id]
#         else:
#             cls_name = "sans_etiquette"
#         save_dir = os.path.join(folder, os.path.basename(cla_filename)[:-4] + "_images_individuelles", cls_name)
#         os.makedirs(save_dir, exist_ok=True)
#         im = last_image[row['image_y']:row['image_y'] + row['image_h'], row['image_x']:row['image_x'] + row['image_w'], :]
#         print(os.path.join(save_dir, run_id + "_{:08d}.png".format(id)))
#         sub_filename = os.path.join(save_dir, run_id + "_{:08d}.png".format(id))
#         skio.imsave(sub_filename, im)
#         df.loc[row_id, 'file'] = os.path.basename(sub_filename)
#     df.to_csv(os.path.join(folder, run_id + "_info.csv"))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Segment flowcam images into individual images sorted by class')
    parser.add_argument("-i", "--input", type=str, help="Input directory containing the flowcam data CSV file and collated images")
    parser.add_argument("-o", "--output", type=str, default=None, required=True, help="Output directory to save images (if not used, images will be saved in a directory alongside the data CSV file)")
    parser.add_argument("-n", "--name", required=True, help="Name of the campaign this sample is from")
    parser.add_argument("-s", "--species", type=str, required=False,  help="XLSX file with the map of class names to true species identifiers")
    args = parser.parse_args()

    print('-' * 80)
    print("Flowcam segmenter")
    print("- species XLSX: {}".format(args.species))
    process_dir(args.input, args.output,  args.name, args.species, True)



"""
Creates and trains a generic network
"""
import time
import datetime
from collections import OrderedDict

from miso.stats.mislabelling import find_and_save_mislabelled
from miso.archive.datasource import DataSource
from miso.archive.generators import *
from miso.utils.wave import *
from miso.training.adaptive_learning_rate import AdaptiveLearningRateScheduler
from miso.training.training_result import TrainingResult
from miso.stats.confusion_matrix import *
from miso.stats.training import *
from archive.augmentation import *
from miso.deploy.saving import freeze, convert_to_inference_mode
from miso.deploy.model_info import ModelInfo
from miso.models.model_factory import *


def train_image_classification_model(params: dict, data_source: DataSource = None):
    K.clear_session()

    config = tf.ConfigProto()
    config.gpu_options.allow_growth = True
    session = tf.Session(config=config)

    intro()

    # Params -----------------------------------------------------------------------------------------------------------
    name = params.get('name')
    description = params.get('description')

    # Network
    cnn_type = params.get('type')

    # Input
    img_size = params.get('img_size')
    if img_size is not None:
        [img_height, img_width, img_channels] = params.get('img_size')
    else:
        img_height = params.get('img_height')
        img_width = params.get('img_width')
        img_channels = params.get('img_channels')

    # Training
    batch_size = params.get('batch_size', 64)
    max_epochs = params.get('max_epochs', 1000)
    alr_epochs = params.get('alr_epochs', 10)
    alr_drops = params.get('alr_drops', 4)

    # Input data
    input_dir = params.get('input_source', None)
    data_min_count = params.get('data_min_count', 40)
    data_split = params.get('data_split', 0.2)
    data_split_offset = params.get('data_split_offset', 0)
    seed = params.get('seed', None)

    # Output
    output_dir = params.get('save_dir')

    # Type
    # - rgb
    # - greyscale
    # - greyscale3
    # - rgbd
    # - greyscaled
    img_type = params.get('img_type', None)
    if img_type is None:
        if img_channels == 3:
            img_type = 'rgb'
        elif img_channels == 1:
            if cnn_type.endswith('tl'):
                img_type = 'greyscale3'
                params['img_channels'] = 3
            else:
                img_type = 'greyscale'
        else:
            raise ValueError("Number of channels must be 1 or 3")
    elif img_type == 'rgbd':
        params['img_channels'] = 4
    elif img_type == 'greyscaled':
        params['img_channels'] = 2
    elif img_type == 'greyscaledm':
        params['img_channels'] = 3

    print('@ Image type: {}'.format(img_type))

    # Data -------------------------------------------------------------------------------------------------------------
    if data_source is None:
        data_source = DataSource()
        data_source.use_mmap = params['use_mmap']
        data_source.set_source(input_dir,
                               data_min_count,
                               mapping=params['class_mapping'],
                               min_count_to_others=params['data_map_others'],
                               mmap_directory=params['mmap_directory'])
        data_source.load_dataset(img_size=(img_height, img_width), img_type=img_type)
    data_source.split(data_split, seed)

    if params['use_class_weights'] is True:
        params['class_weights'] = data_source.get_class_weights()
        print("@ Class weights are {}".format(params['class_weights']))
    else:
        params['class_weights'] = None
    params['num_classes'] = data_source.num_classes

    if cnn_type.endswith('tl'):
        start = time.time()

        # Generate vectors
        model_head = generate_tl_head(params)
        print("@ Calculating train vectors")
        t = time.time()
        train_vector = model_head.predict(data_source.train_images)
        print("! {}s elapsed".format(time.time() - t))
        print("@ Calculating test vectors")
        t = time.time()
        test_vector = model_head.predict(data_source.test_images)
        print("! {}s elapsed".format(time.time() - t))
        # Clear
        K.clear_session()

        config = tf.ConfigProto()
        config.gpu_options.allow_growth = True
        session = tf.Session(config=config)

        data_source.train_vectors = train_vector
        data_source.test_vectors = test_vector

        # Augmentation -------------------------------------------------------------------------------------------------
        # No augmentation as we pre-calculate vectors

        # Generator ----------------------------------------------------------------------------------------------------
        # No generator needed

        # Model --------------------------------------------------------------------------------------------------------
        print("@ Generating tail")
        # Get  tail
        model_tail = generate_tl_tail(params, [train_vector.shape[1], ])
        model_tail.compile(optimizer='adam', loss='categorical_crossentropy', metrics=['accuracy'])

        # Generator ----------------------------------------------------------------------------------------------------
        train_gen = tf_vector_generator(train_vector,
                                        data_source.train_onehots,
                                        batch_size)
        test_gen = tf_vector_generator(test_vector,
                                       data_source.test_onehots,
                                       batch_size)

        # Training -----------------------------------------------------------------------------------------------------
        alr_cb = AdaptiveLearningRateScheduler(nb_epochs=alr_epochs,
                                               nb_drops=alr_drops)
        print("@ Training")
        if data_split > 0:
            validation_data = test_gen
        else:
            validation_data = None
        # log_dir = "C:\\logs\\profile\\" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        # tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1, profile_batch=3)
        history = model_tail.fit_generator(
            train_gen,
            steps_per_epoch=math.ceil(len(train_vector) // batch_size),
            validation_data=validation_data,
            validation_steps=math.ceil(len(test_vector) // batch_size),
            epochs=max_epochs,
            verbose=0,
            shuffle=False,
            max_queue_size=1,
            class_weight=params['class_weights'],
            callbacks=[alr_cb])
        end = time.time()
        training_time = end - start
        print("@ Training time: {}s".format(training_time))
        time.sleep(3)

        # Generator ----------------------------------------------------------------------------------------------------
        # Now we be tricky and join the trained dense layers to the resnet model to create a model that accepts images
        # as input
        model_head = generate_tl_head(params)
        outputs = model_tail(model_head.output)
        model = Model(model_head.input, outputs)
        model.summary()

        # Vector -------------------------------------------------------------------------------------------------------
        vector_model = generate_vector(model, params)

    else:
        # Model --------------------------------------------------------------------------------------------------------
        print("@ Generating model")
        start = time.time()
        model = generate(params)

        # Augmentation -------------------------------------------------------------------------------------------------
        if params['aug_rotation'] is True:
            rotation_range = [0, 360]
        else:
            rotation_range = None

        def augment(x):
            return augmentation_complete(x,
                                         rotation=rotation_range,
                                         gain=params['aug_gain'],
                                         gamma=params['aug_gamma'],
                                         zoom=params['aug_zoom'],
                                         gaussian_noise=params['aug_gaussian_noise'],
                                         bias=params['aug_bias'])

        if params['use_augmentation'] is True:
            augment_fn = augment
        else:
            augment_fn = None

        # Generator ----------------------------------------------------------------------------------------------------
        train_gen = tf_augmented_image_generator(data_source.train_images,
                                                 data_source.train_onehots,
                                                 batch_size,
                                                 augment_fn)
        test_gen = image_generator(data_source.test_images,
                                   data_source.test_onehots,
                                   batch_size)

        # Training -----------------------------------------------------------------------------------------------------

        alr_cb = AdaptiveLearningRateScheduler(nb_epochs=alr_epochs,
                                               nb_drops=alr_drops)
        print("@ Training")
        if data_split > 0:
            validation_data = test_gen
        else:
            validation_data = None
        history = model.fit_generator(
            train_gen,
            steps_per_epoch=math.ceil(len(data_source.train_images) // batch_size),
            validation_data=validation_data,
            validation_steps=math.ceil(len(data_source.test_images) // batch_size),
            epochs=max_epochs,
            verbose=0,
            shuffle=False,
            max_queue_size=1,
            class_weight=params['class_weights'],
            callbacks=[alr_cb])
        end = time.time()
        training_time = end - start
        print("@ Training time: {}s".format(training_time))
        time.sleep(3)

        # Vector -------------------------------------------------------------------------------------------------------
        vector_model = generate_vector(model, params)

    # Graphs -----------------------------------------------------------------------------------------------------------
    print("@ Generating results")
    if data_split > 0:
        # Calculate test set scores
        y_true = data_source.test_cls
        y_prob = model.predict(data_source.test_images)
        y_pred = y_prob.argmax(axis=1)
    else:
        y_true = np.asarray([])
        y_prob = np.asarray([])
        y_pred = np.asarray([])

    # Inference time
    max_count = np.min([1000, len(data_source.images)])
    to_predict = np.copy(data_source.images[0:max_count])

    inf_times = []
    for i in range(3):
        start = time.time()
        model.predict(to_predict)
        end = time.time()
        diff = (end - start) / max_count * 1000
        inf_times.append(diff)
        print("@ Calculating inference time {}/10: {:.3f}ms".format(i + 1, diff))
    inference_time = np.median(inf_times)

    # Store results
    result = TrainingResult(params,
                            history,
                            y_true,
                            y_pred,
                            y_prob,
                            data_source.cls_labels,
                            training_time,
                            inference_time)

    # Save the results
    now = datetime.datetime.now()
    save_dir = os.path.join(output_dir, "{0}_{1:%Y%m%d-%H%M%S}".format(name, now))
    os.makedirs(save_dir, exist_ok=True)

    # Plot the graphs
    # plot_model(model, to_file=os.path.join(save_dir, "model_plot.pdf"), show_shapes=True)
    if data_split > 0:
        plot_loss_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "loss_vs_epoch.pdf"))
        plot_accuracy_vs_epochs(history)
        plt.savefig(os.path.join(save_dir, "accuracy_vs_epoch.pdf"))
        plot_confusion_accuracy_matrix(y_true, y_pred, data_source.cls_labels)
        plt.savefig(os.path.join(save_dir, "confusion_matrix.pdf"))
        plt.close('all')

    if params['save_mislabeled'] is True:
        print("@ Estimating mislabeled")
        vectors = vector_model.predict(data_source.images)
        find_and_save_mislabelled(data_source.images,
                                  vectors,
                                  data_source.cls,
                                  data_source.cls_labels,
                                  data_source.get_short_filenames(),
                                  save_dir,
                                  11)

    # Save model -------------------------------------------------------------------------------------------------------
    print("@ Saving model")
    # Convert if necessary to fix TF batch normalisation issues
    model = convert_to_inference_mode(model, lambda: generate(params))
    vector_model = generate_vector(model, params)

    # Generate description
    if description is None:
        description = "{}: {} model trained on data from {} ({} images in {} classes).\n" \
                      "Accuracy: {:.1f} (P: {:.1f}, R: {:.1f}, F1 {:.1f})".format(
            name,
            cnn_type,
            input_dir,
            len(data_source.data_df),
            len(data_source.cls_labels),
            result.accuracy * 100,
            result.mean_precision() * 100,
            result.mean_recall() * 100,
            result.mean_f1_score() * 100)

    # Create model info with all the parameters
    inputs = OrderedDict()
    inputs["image"] = model.inputs[0]
    outputs = OrderedDict()
    outputs["pred"] = model.outputs[0]
    outputs["vector"] = vector_model.outputs[0]
    info = ModelInfo(name,
                     description,
                     cnn_type,
                     now,
                     "frozen_model.pb",
                     params,
                     inputs,
                     outputs,
                     input_dir,
                     data_source.cls_labels,
                     data_source.cls_counts,
                     "rescale",
                     [255, 0, 1],
                     result.accuracy,
                     result.precision,
                     result.recall,
                     result.f1_score,
                     result.support,
                     result.epochs[-1],
                     training_time,
                     params['data_split'],
                     inference_time)

    # Freeze and save graph
    if params['save_model'] is not None:
        freeze(model,
               os.path.join(save_dir, "model"),
               info)

    # Save info
    info.save(os.path.join(save_dir, "model", "network_info.xml"))

    print("@ Deleting temporary files")
    data_source.delete_memmap_files(del_split=True, del_source=params['delete_mmap_files'])

    wave()

    print("@ Complete")
    return model, vector_model, data_source, result

# tensorboard_cb = TensorBoard(log_dir='./tensorboard',
#                              histogram_freq=0,
#                              batch_size=32,
#                              write_graph=True,
#                              write_grads=False,
#                              write_images=False,
#                              embeddings_freq=0,
#                              embeddings_layer_names=None,
#                              embeddings_metadata=None,
#                              embeddings_data=None,
#                              update_freq='epoch')



import os
import numpy as np
import matplotlib.pyplot as plt
from sklearn.neighbors.nearest_centroid import NearestCentroid
from sklearn.preprocessing import normalize
from sklearn.metrics import pairwise_distances_argmin_min


def plot_most_representative(images, vectors, cls, cls_labels, output_dir):
    # Normalise vectors
    vectors = normalize(vectors, axis=1)
    # Find centroids
    clf = NearestCentroid()
    clf.fit(vectors, cls)
    # Get nearest
    im_idx, dist = pairwise_distances_argmin_min(clf.centroids_, vectors)

    for i, idx in enumerate(im_idx):
        plt.clf()
        plt.imshow(images[idx])
        plt.axis('off')
        plt.title('{}  -  {}'.format(i, cls_labels[i]))
        plt.tight_layout()
        plt.show()
        output_file = os.path.join(output_dir, 'archetypes', '{}_{}.png'.format(i, cls_labels[i]))
        plt.imsave(output_file)




# import os
#
# import tempfile
# from miso.utils import singleton
#
#
# def try_get_lock():
#     try:
#         fn = os.path.join(tempfile.gettempdir(), "miso.lock")
#         with open(fn, 'w') as fh:
#             fh.write('miso')
#
#
#         try:
#             os.chmod(fn, 0o777)
#         except OSError:
#         pass
#
#     lock = singleton.SingleInstance(lockfile=fn)
#     print()
#     train_image_classification_model(tp)
#     done = True
#     except singleton.SingleInstanceException:
#     print(\{}: Another
#     script is already
#     running, trying
#     again in 10
#     seconds.({}
#     s
#     waiting)\\r\.format(datetime.now(), np.round(time.time() - start)), end = '')
#     time.sleep(10);



import datetime
from typing import List, Dict
from lxml import etree as ET
import os
import numpy as np
from collections import OrderedDict

from miso.training.parameters import MisoParameters


class ModelInfo:
    def __init__(self,
                 name: str,
                 description: str,
                 type: str,
                 date: datetime.datetime,
                 protobuf: str,
                 params: MisoParameters,
                 inputs: OrderedDict,
                 outputs: OrderedDict,
                 data_source_name: str,
                 labels: List[str],
                 counts: List[int],
                 prepro_name: str,
                 prepro_params: List,
                 accuracy: float,
                 precision: float,
                 recall: float,
                 f1score: float,
                 support: float,
                 training_epochs: int,
                 training_time: float,
                 training_split: float,
                 inference_time_per_image: float):

        self.name = name
        self.description = description
        self.type = type
        self.date = date
        self.params = params
        self.inputs = inputs
        self.outputs = outputs
        self.data_source_name = data_source_name
        self.labels = labels
        self.counts = counts
        self.prepro_name = prepro_name
        self.prepro_params = prepro_params
        self.protobuf = protobuf
        self.accuracy = accuracy
        self.precision = precision
        self.recall = recall
        self.f1score = f1score
        self.support = support
        self.training_epochs = training_epochs
        self.training_time = training_time
        self.training_split = training_split
        self.inference_time_per_image = inference_time_per_image
        self.version = "2.1"

    def save(self, filename):
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        f = open(filename, 'wb')
        f.write(self.to_xml())
        f.close()

    def _get_op_name(self, tensor):
        try:
            return tensor.op.name
        except:
            return tensor.name

    def to_xml(self):
        root = ET.Element("network", version=self.version)
        ET.SubElement(root, "name").text = self.name
        ET.SubElement(root, "description").text = self.description
        ET.SubElement(root, "type").text = self.type
        ET.SubElement(root, "date").text = "{0:%Y-%m-%d_%H%M%S}".format(self.date)
        ET.SubElement(root, "protobuf").text = self.protobuf

        # parent_node = ET.SubElement(root, "params")
        # for key, value in self.params.asdict().items():
        #     ET.SubElement(parent_node, key).text = str(value)

        parent_node = ET.SubElement(root, "inputs")
        for name, tensor in self.inputs.items():
            node = ET.SubElement(parent_node, "input")
            ET.SubElement(node, "name").text = name
            ET.SubElement(node, "operation").text = self._get_op_name(tensor)
            ET.SubElement(node, "height").text = str(tensor.shape[1])
            if len(tensor.shape) > 2:
                ET.SubElement(node, "width").text = str(tensor.shape[2])
            else:
                ET.SubElement(node, "width").text = "0"
            if len(tensor.shape) > 3:
                ET.SubElement(node, "channels").text = str(tensor.shape[3])
            else:
                ET.SubElement(node, "channels").text = "0"

        parent_node = ET.SubElement(root, "outputs")
        for name, tensor in self.outputs.items():
            node = ET.SubElement(parent_node, "output")
            ET.SubElement(node, "name").text = name
            ET.SubElement(node, "operation").text = self._get_op_name(tensor)
            ET.SubElement(node, "height").text = str(tensor.shape[1])
            if len(tensor.shape) > 2:
                ET.SubElement(node, "width").text = str(tensor.shape[2])
            else:
                ET.SubElement(node, "width").text = "0"
            if len(tensor.shape) > 3:
                ET.SubElement(node, "channels").text = str(tensor.shape[3])
            else:
                ET.SubElement(node, "channels").text = "0"

        ET.SubElement(root, "source_data").text = str(self.data_source_name)
        ET.SubElement(root, "source_size").text = str(np.sum(self.counts))
        parent_node = ET.SubElement(root, "labels")
        for idx, value in enumerate(self.labels):
            node = ET.SubElement(parent_node, "label")
            ET.SubElement(node, "code").text = value
            ET.SubElement(node, "count").text = str(self.counts[idx])
            ET.SubElement(node, "precision").text = str(self.precision[idx])
            ET.SubElement(node, "recall").text = str(self.recall[idx])
            ET.SubElement(node, "f1score").text = str(self.f1score[idx])
            ET.SubElement(node, "support").text = str(self.support[idx])

        parent_node = ET.SubElement(root, "prepro")
        ET.SubElement(parent_node, "name").text = self.prepro_name
        parent_node = ET.SubElement(parent_node, "params")
        for idx, value in enumerate(self.prepro_params):
            ET.SubElement(parent_node, "param").text = str(value)

        ET.SubElement(root, "accuracy").text = str(self.accuracy)
        ET.SubElement(root, "precision").text = str(np.mean(self.precision))
        ET.SubElement(root, "recall").text = str(np.mean(self.recall))
        ET.SubElement(root, "f1score").text = str(np.mean(self.f1score))

        parent_node = ET.SubElement(root, "load")
        ET.SubElement(parent_node, "training_epochs").text = str(self.training_epochs)
        ET.SubElement(parent_node, "training_time").text = str(self.training_time)
        ET.SubElement(parent_node, "training_split").text = str(self.training_split)
        ET.SubElement(parent_node, "training_time_per_image").text = str(
            self.training_time / self.training_epochs / (np.sum(self.counts) * (1 - self.training_split)))
        ET.SubElement(parent_node, "inference_time_per_image").text = str(np.mean(self.inference_time_per_image))

        return ET.tostring(root, pretty_print=True)

    @staticmethod
    def from_xml(xml_data_or_path):
        if os.path.exists(xml_data_or_path):
            tree = ET.parse(xml_data_or_path)
            root = tree.getroot()
        else:
            root = ET.fromstring(xml_data_or_path)

        name = root.findtext('name')
        description = root.findtext('description')
        model_type = root.findtext('type')
        date_str = root.findtext('date')
        date = datetime.datetime.strptime(date_str, "%Y-%m-%d_%H%M%S")
        protobuf = root.findtext('protobuf')

        inputs = []
        for inp in root.find('inputs').findall('input'):
            d = {}
            d["name"] = inp.findtext('name')
            d["operation"] = inp.findtext('operation')
            d["height"] = inp.findtext('height')
            d["width"] = inp.findtext('width')
            d["channels"] = inp.findtext('channels')
            inputs.append(d)

        outputs = []
        for out in root.find('outputs').findall('output'):
            d = {}
            d["name"] = out.findtext('name')
            d["operation"] = out.findtext('operation')
            d["height"] = out.findtext('height')
            d["width"] = out.findtext('width')
            d["channels"] = out.findtext('channels')
            outputs.append(d)

        labels = [label.findtext('code') for label in root.findall('.//labels/label')]

        return ModelInfo(name,
                         description,
                         model_type,
                         date,
                         protobuf,
                         None,
                         inputs,
                         outputs,
                         None,
                         labels,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None,
                         None)



import tensorflow as tf


def ASoftmax(x, y, l, num_cls, m=2, name='asoftmax'):
    '''
    x: B x D - data
    y: B x 1 - label
    l: 1 - lambda
    '''
    def wrapper(x, y):
        x = x.get_shape()
        w = tf.get_variable("asoftmax/W", [xs[1], num_cls], dtype=tf.float32,
                            initializer=tf.contrib.layers.xavier_initializer())
        eps = 1e-8
        xw = tf.matmul(x, w)
        if m == 0:
            return xw, tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=y, logits=xw))

        w_norm = tf.norm(w, axis=0) + eps
        logits = xw / w_norm

        if y is None:
            return logits, None

        ordinal = tf.constant(list(range(0, xs[0])), tf.int64)
        ordinal_y = tf.stack([ordinal, y], axis=1)

        x_norm = tf.norm(x, axis=1) + eps
        sel_logits = tf.gather_nd(logits, ordinal_y)
        cos_th = tf.divide(sel_logits, x_norm)
        if m == 1:
            loss = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=y, logits=logits))
        else:
            if m == 2:
                cos_sign = tf.sign(cos_th)
                res = 2 * tf.multiply(tf.sign(cos_th), tf.square(cos_th)) - 1
            elif m == 4:
                cos_th2 = tf.square(cos_th)
                cos_th4 = tf.pow(cos_th, 4)
                sign0 = tf.sign(cos_th)
                sign3 = tf.multiply(tf.sign(2 * cos_th2 - 1), sign0)
                sign4 = 2 * sign0 + sign3 - 3
                res = sign3 * (8 * cos_th4 - 8 * cos_th2 + 1) + sign4
            else:
                raise ValueError('unsupported value of m')

            scaled_logits = tf.multiply(res, x_norm)
            f = 1.0 / (1.0 + l)
            ff = 1.0 - f
            comb_logits_diff = tf.add(logits, tf.scatter_nd(ordinal_y, tf.subtract(scaled_logits, sel_logits), logits.get_shape()))
            updated_logits = ff * logits + f * comb_logits_diff

            loss = tf.reduce_mean(tf.nn.sparse_softmax_cross_entropy_with_logits(labels=y, logits=updated_logits))

        return logits, loss



import numpy as np
from skimage import io as skio, transform as skt, color as skc


def load_image(filename, img_size=None, img_type='rgb'):
    """
    Loads an image, converting it if necessary
    :param filename: Filename of image to load
    :param img_size: Size of image, e.g. (224, 224). If None, the image dimensions are preserved
    :param img_type: 'rgb' (colour) or 'k' (greyscale)
    :return:
    """
    # Colour
    if img_type == 'rgb':
        im = skio.imread(filename)
        # im = np.asarray(im, dtype=np.float)
        # If it was a single channel image, make into 3-channel
        if im.ndim == 2:
            im = np.repeat(im[..., np.newaxis], repeats=3, axis=-1)
    # Greyscale
    elif img_type == 'k' or img_type == 'greyscale':
        im = skio.imread(filename, as_gray=True)
        im = im[..., np.newaxis]
    else:
        raise ValueError("img_type must be 'rgb' or 'k'")
    # Resize and pad
    if img_size is not None:
        im = resize_and_pad_image(im, img_size)
    return im


def resize_and_pad_image(im, img_size):
    # Add last dim if greyscale
    if np.ndim(im) == 2:
        im = im[..., np.newaxis]
    # Get the ratio of width to height for each
    current_whratio = im.shape[1] / im.shape[0]
    desired_whratio = img_size[1] / img_size[0]
    # Check if the image has roughly the same ratio, else pad it
    if np.round(im.shape[0] * desired_whratio) != im.shape[1]:
        height = im.shape[0]
        width = im.shape[1]
        # Desired shape is wider than current one
        if desired_whratio > current_whratio:
            half = np.round(height * desired_whratio)
            height_pad_start = 0
            height_pad_end = 0
            width_pad_start = int(abs(np.floor((width - half) / 2)))
            width_pad_end = int(abs(np.ceil((width - half) / 2)))
        # Desired shape is taller than current
        else:
            half = np.round(width / desired_whratio)
            height_pad_start = int(abs(np.floor((height - half) / 2)))
            height_pad_end = int(abs(np.ceil((height - half) / 2)))
            width_pad_start = 0
            width_pad_end = 0
        # Constant value to pad with
        consts = [np.median(np.concatenate((im[0, :, i], im[-1, :, i], im[:, 0, i], im[:, -1, i]))) for i in range(im.shape[2])]
        # Pad
        im = np.stack(
            [np.pad(im[:, :, c],
                    ((height_pad_start, height_pad_end), (width_pad_start, width_pad_end)),
                    mode='constant',
                    constant_values=consts[c])
             for c in range(im.shape[2])], axis=2)
        # # Revert if was greyscale
        # if im.shape[2] == 1:
        #     im = im[..., 0]
    # Resize
    if im.dtype == np.uint8:
        im = im / 255
    if im.shape[0] != img_size[0] or im.shape[1] != img_size[1]:
        im = skt.resize(im, img_size)
    return im


def to_channels(im, out_channels, to_greyscale=False):
    # Number of channels in image
    if np.ndim(im) == 3:
        img_channels = im.shape[2]
    else:
        im = im[..., np.newaxis]
        img_channels = 1
    # Convert to greyscale if needed:
    if img_channels == 3 and to_greyscale is True:
        im = skc.rgb2gray(im)[..., np.newaxis]
        img_channels = 1
    # Repeat if needed
    if img_channels == 1 and out_channels == 3:
        im = np.repeat(im, 3, axis=-1)
    return im


def resize_transform(im, shape, to_greyscale=False):
    im = to_channels(im, shape[2], to_greyscale)
    if im.shape[0] == shape[0] and im.shape[1] == shape[1]:
        return im
    im = skt.resize(im, shape)
    return (im * 255).astype(np.uint8)


def resize_with_pad_transform(im, shape, to_greyscale=False):
    im = to_channels(im, shape[2], to_greyscale)
    if im.shape[0] == shape[0] and im.shape[1] == shape[1]:
        return im
    im = resize_and_pad_image(im, shape)
    return (im * 255).astype(np.uint8)


def null_transform(im, args):
    return im



import tensorflow as tf
import numpy as np


def predict_in_batches(model, generator):
    results = []
    try:
        for x, y in iter(generator):
            results.append(model.predict(x))
    except tf.errors.OutOfRangeError:
        pass
    results = np.concatenate(results, axis=0)
    return results


from typing import NamedTuple

from sklearn.model_selection import train_test_split
from keras.utils import to_categorical
import numpy as np
from scipy.stats.mstats import gmean

from miso.data.filenames_dataset import FilenamesDataset
from miso.data.image_dataset import ImageDataset
from miso.data.tf_generator import TFGenerator


class TrainingDataset(object):
    def __init__(self,
                 source,
                 img_size=(224, 224, 3),
                 img_type='rgb',
                 min_count=0,
                 map_others=False,
                 train_split=None,
                 test_split=0.2,
                 random_seed=0,
                 memmap_directory=None):
        if len(img_size) != 3:
            raise ValueError("img_size must be in format [height, width, num_channels]")
        self.source = source
        self.img_size = img_size
        self.img_type = img_type
        self.min_count = min_count
        self.map_others = map_others
        self.test_split = test_split
        self.train_split = train_split
        self.random_seed = random_seed
        self.memmap_directory = memmap_directory

        self.filenames: FilenamesDataset = None
        self.images: ImageDataset = None
        self.train_idx = None
        self.test_idx = None
        self.cls = None
        self.cls_onehot = None
        self.cls_labels = None
        self.num_classes = None
        self.class_weights = None

    def get_class_weights(self):
        count = np.bincount(self.data_df['cls'])
        weights = gmean(count) / count
        weights[weights < 0.1] = 0.1
        weights[weights > 10] = 10
        return weights

    def load(self):
        # Get filenames
        fs = FilenamesDataset(self.source, has_classes=True)
        fs.load(self.min_count, self.map_others)
        self.filenames = fs
        self.cls = self.filenames.cls
        self.cls_labels = fs.cls_labels
        self.num_classes = fs.num_classes

        # Create one hot
        self.cls_onehot = to_categorical(fs.cls)

        # Class weights
        weights = gmean(fs.cls_counts) / fs.cls_counts
        weights[weights < 0.1] = 0.1
        weights[weights > 10] = 10
        self.class_weights = weights

        # Create split
        if self.test_split > 0:
            self.train_idx, self.test_idx = train_test_split(np.arange(len(self.filenames.filenames)),
                                                             stratify=self.cls,
                                                             test_size=self.test_split,
                                                             train_size=self.train_split,
                                                             random_state=self.random_seed)
        else:
            self.train_idx = np.arange(len(self.filenames.filenames))
            self.test_idx = []

        # Load images
        to_greyscale = False
        if self.img_type == 'k' or self.img_type == 'greyscale':
            to_greyscale = True
        # print(self.img_size)
        self.images = ImageDataset(self.filenames.filenames,
                                   self.cls_onehot,
                                   transform_fn='resize_with_pad',
                                   transform_args=[self.img_size, to_greyscale],
                                   memmap_directory=self.memmap_directory)
        self.images.load()

    def train_generator(self, batch_size=32, shuffle=True, one_shot=False, undersample=False, map_fn=TFGenerator.map_fn_divide_255):
        return self.images.create_generator(batch_size, self.train_idx, map_fn=map_fn, shuffle=shuffle, one_shot=one_shot, undersample=undersample)

    def test_generator(self, batch_size=32, shuffle=True, one_shot=False, undersample=False, map_fn=TFGenerator.map_fn_divide_255):
        return self.images.create_generator(batch_size, self.test_idx, map_fn=map_fn, shuffle=shuffle, one_shot=one_shot, undersample=undersample)

    def release(self):
        self.images.release()


if __name__ == "__main__":
    source = "/Users/chaos/Documents/Development/Data/Modern_Coretop_Source/project.xml"
    import matplotlib.pyplot as plt
    # XML
    ts = TrainingDataset(source,
                         img_size=[224, 224, 1],
                         img_type='k',
                         min_count=10,
                         map_others=False,
                         test_split=0.2,
                         random_seed=0,
                         memmap_directory=None)
    ts.load(32)
    im = ts.images.data[0]
    plt.imshow(im)
    plt.title("shape: {}, max: {}, min: {}".format(im.shape, im.max(), im.min()))
    plt.show()

    ts = TrainingDataset(source,
                         img_size=[224, 224, 3],
                         img_type='k',
                         min_count=10,
                         map_others=False,
                         test_split=0.2,
                         random_seed=0,
                         memmap_directory=None)
    ts.load(32)
    im = ts.images.data[0]
    plt.imshow(im)
    plt.title("shape: {}, max: {}, min: {}".format(im.shape, im.max(), im.min()))
    plt.show()

    ts = TrainingDataset(source,
                         img_size=[224, 224, 3],
                         img_type='rgb',
                         min_count=10,
                         map_others=False,
                         test_split=0.2,
                         random_seed=0,
                         memmap_directory=None)
    ts.load(32)
    im = ts.images.data[0]
    plt.imshow(im)
    plt.title("shape: {}, max: {}, min: {}".format(im.shape, im.max(), im.min()))
    plt.show()



import os
import time

import skimage
import tensorflow as tf
import numpy as np

from miso.data.tf_generator import TFGenerator
from miso.data.training_dataset import TrainingDataset
from miso.models.factory import generate
from miso.training.adaptive_learning_rate import AdaptiveLearningRateScheduler
from miso.training.parameters import MisoParameters
from miso.training.augmentation import aug_all_fn


def save_images(images, output_dir):
    os.makedirs(output_dir, exist_ok=True)
    for t_idx, im in enumerate(images):
        im = im.numpy() * 255
        im[im > 255] = 255
        if np.ndim(im) == 2:
            im = np.repeat(im[:, :, np.newaxis], 3, axis=-1)
        elif im.shape[-1] == 1:
            im = np.repeat(im, 3, axis=-1)
        skimage.io.imsave(
            os.path.join(output_dir, "{:03d}.jpg".format(t_idx)),
            im.astype(np.uint8),
        )


def train_full_network(tp: MisoParameters, ds: TrainingDataset, save_dir: str):
    # Tensorflow version
    tf_version = int(tf.__version__[0])

    print("-" * 80)
    print("Full network training")
    start = time.time()

    # Generate model
    model = generate(tp)
    # model.summary()

    # Augmentation
    if tp.augmentation.rotation is True:
        tp.augmentation.rotation = [0, 360]
    elif tp.augmentation.rotation is False:
        tp.augmentation.rotation = None
    if tp.training.use_augmentation is True:
        print("- using augmentation")
        augment_fn = aug_all_fn(
            rotation=tp.augmentation.rotation,
            gain=tp.augmentation.gain,
            gamma=tp.augmentation.gamma,
            zoom=tp.augmentation.zoom,
            gaussian_noise=tp.augmentation.gaussian_noise,
            bias=tp.augmentation.bias,
            random_crop=tp.augmentation.random_crop,
            divide=255,
        )
    else:
        print("- NOT using augmentation")
        augment_fn = TFGenerator.map_fn_divide_255

    # Learning rate scheduler
    alr_cb = AdaptiveLearningRateScheduler(
        nb_epochs=tp.training.alr_epochs, nb_drops=tp.training.alr_drops, verbose=1
    )

    # Training generator
    train_gen = ds.train_generator(
        batch_size=tp.training.batch_size,
        map_fn=augment_fn,
        undersample=tp.training.use_class_undersampling,
    )

    # Save example of training data
    print("- saving example training batch")
    training_examples_dir = os.path.join(save_dir, "examples", "train")
    images, labels = next(iter(train_gen.create()))
    save_images(images, training_examples_dir)

    # Validation generator
    if tf_version == 2:
        val_one_shot = True
    else:
        # One repeat for validation for TF1 otherwise we get end of dataset errors
        val_one_shot = False
    if tp.dataset.val_split > 0:
        # Maximum 8 in batch otherwise validation results jump around a bit because
        val_gen = ds.test_generator(
            min(tp.training.batch_size, 16), shuffle=False, one_shot=val_one_shot
        )
        val_data = val_gen.create()
        val_steps = len(val_gen)

        # Save example of validation data
        print("- saving example validation batch")
        training_examples_dir = os.path.join(save_dir, "examples", "val")
        images, labels = next(iter(val_gen.create()))
        save_images(images, training_examples_dir)

    else:
        val_data = None
        val_steps = None

    # Class weights
    if (tp.training.use_class_weights is True and tp.training.use_class_undersampling is False):
        class_weights = ds.class_weights
        print("- class weights: {}".format(class_weights))
        if tf_version == 2:
            class_weights = dict(enumerate(class_weights))
    else:
        class_weights = None
    if tp.training.use_class_undersampling:
        print("- class balancing using random under sampling")

    # Train the model
    if tf_version == 2:
        train_fn = model.fit
    else:
        train_fn = model.fit_generator
    history = train_fn(
        train_gen.create(),
        steps_per_epoch=len(train_gen),
        validation_data=val_data,
        validation_steps=val_steps,
        epochs=tp.training.max_epochs,
        verbose=0,
        shuffle=False,
        max_queue_size=1,
        class_weight=class_weights,
        callbacks=[alr_cb],
    )

    # Elapsed time
    end = time.time()
    training_time = end - start
    print()
    print("Total training time: {}s".format(training_time))
    history.history["training_time"] = training_time

    return model, history


import matplotlib.pyplot as plt
import numpy as np
from sklearn.metrics import precision_recall_fscore_support, accuracy_score


def plot_precision_recall(y_true,
                          y_pred,
                          cls_labels,
                          fig_size=(6, 4),
                          rotate_labels=90,
                          show=False):
    p, r, f1, s = precision_recall_fscore_support(y_true, y_pred, labels=range(len(cls_labels)))
    plt.figure(facecolor="white", figsize=fig_size)
    ax = plt.subplot(111)
    ax.set_axisbelow(True)
    ax.yaxis.grid(color='gray', linestyle='dashed')
    w = 0.2
    x = np.arange(len(p))
    class_precision = p * 100
    class_recall = r * 100
    ax = plt.gca()
    h1 = ax.bar(x - w, class_recall, width=2 * w, align='center')
    h2 = ax.bar(x + w, class_precision, width=2 * w, align='center')
    h3 = ax.plot(x, f1*100, 'k.')
    plt.xticks(x, cls_labels, rotation=rotate_labels)
    plt.ylim(0, 100)
    plt.xlim(x[0]-0.5, x[-1]+0.5)
    plt.legend((h1[0], h2[0], h3[0]),
               ['recall: {0:.1f}%'.format(np.mean(r) * 100),
                'precision: {0:.1f}%'.format(100 * np.mean(p)),
                'f1: {0:.1f}%'.format(100 * np.mean(f1))],
               loc=(0, 1.04),
               ncol=3)
    plt.xlabel('Accuracy {0:.1f}%'.format(accuracy_score(y_true, y_pred)*100))
    plt.tight_layout()
    # plt.subplots_adjust(top=0.85)
    if show:
        plt.show()



import tensorflow as tf

def batch_instance_norm(x, scope='batch_instance_norm'):
    with tf.variable_scope(scope):
        ch = x.shape[-1]
        eps = 1e-5

        batch_mean, batch_sigma = tf.nn.moments(x, axes=[0, 1, 2], keep_dims=True)
        x_batch = (x - batch_mean) / (tf.sqrt(batch_sigma + eps))

        ins_mean, ins_sigma = tf.nn.moments(x, axes=[1, 2], keep_dims=True)
        x_ins = (x - ins_mean) / (tf.sqrt(ins_sigma + eps))

        rho = tf.get_variable("rho", [ch], initializer=tf.constant_initializer(1.0), constraint=lambda x: tf.clip_by_value(x, clip_value_min=0.0, clip_value_max=1.0))
        gamma = tf.get_variable("gamma", [ch], initializer=tf.constant_initializer(1.0))
        beta = tf.get_variable("beta", [ch], initializer=tf.constant_initializer(0.0))

        x_hat = rho * x_batch + (1 - rho) * x_ins
        x_hat = x_hat * gamma + beta

        return x_hat

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer

class BatchInstanceNormalisation(Layer):

    def __init__(self, **kwargs):
        super(BatchInstanceNormalisation, self).__init__(**kwargs)

    def build(self, input_shape):
        # Create a trainable weight variable for this layer.
        ch = input_shape[-1]
        self.rho = self.add_weight(name='rho',
                                   shape=[ch],
                                   trainable=True,
                                   initializer=tf.constant_initializer(1.0),
                                   constraint=lambda x: tf.clip_by_value(x, clip_value_min=0.0, clip_value_max=1.0))
        self.gamma = self.add_weight(name='gamma',
                                   shape=[ch],
                                   trainable=True,
                                   initializer=tf.constant_initializer(1.0),
                                   constraint=lambda x: tf.clip_by_value(x, clip_value_min=0.0, clip_value_max=1.0))
        self.beta = self.add_weight(name='beta',
                                   shape=[ch],
                                   trainable=True,
                                   initializer=tf.constant_initializer(0.0),
                                   constraint=lambda x: tf.clip_by_value(x, clip_value_min=0.0, clip_value_max=1.0))

        super(BatchInstanceNormalisation, self).build(input_shape)  # Be sure to call this at the end

    def call(self, x):
        eps = 1e-5

        batch_mean, batch_sigma = tf.nn.moments(x, axes=[0, 1, 2], keep_dims=True)
        x_batch = (x - batch_mean) / (tf.sqrt(batch_sigma + eps))

        ins_mean, ins_sigma = tf.nn.moments(x, axes=[1, 2], keep_dims=True)
        x_ins = (x - ins_mean) / (tf.sqrt(ins_sigma + eps))

        x_hat = self.rho * x_batch + (1 - self.rho) * x_ins
        x_hat = x_hat * self.gamma + self.beta

        return x_hat


import numpy as np
import tensorflow as tf
import tensorflow.keras.backend as K
from imblearn.under_sampling import RandomUnderSampler


class TFGenerator(object):
    def __init__(self,
                 data,
                 cls=None,
                 idxs=None,
                 batch_size=32,
                 shuffle=True,
                 prefetch=4,
                 map_fn=None,
                 one_shot=False,
                 undersample=False,
                 data_dtype=tf.float32,
                 labels_dtype=tf.float32):
        """
        Class to create a tf.data.Dataset given a set of data and associated labels.
        Use the create() function to return the dataset
        :param data: Input data
        :param cls: Input class data (e.g. onehot vector or index), can be none
        :param idxs: Indices of the data to use, if None all data are used
        :param batch_size: Batch size for training / inference
        :param shuffle: Whether to shuffle the data each time
        :param prefetch: How many batches to prefetch
        :param map_fn: Function applied to the data when creating a batch. Must take a tensor as input
        :param one_shot: If True, dataset will only iterate through the data once. (Use for validation / inference etc)
        """
        self.data = data
        self.data_dtype = data_dtype
        if isinstance(cls, list):
            cls = np.asarray(cls)
        self.labels = cls
        self.labels_dtype = labels_dtype
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.prefetch = prefetch
        self.map_fn = map_fn
        self.one_shot = one_shot
        self.undersample = undersample
        if idxs is None:
            self.idxs = np.arange(len(data))
        else:
            self.idxs = idxs.copy()
        self.on_epoch_end()

    def __len__(self):
        if self.one_shot:
            return int(np.ceil(len(self.idxs) / self.batch_size))
        else:
            return int(np.floor(len(self.idxs) / self.batch_size))

    def on_epoch_end(self):
        if self.undersample:
            rus = RandomUnderSampler()
            if isinstance(self.labels[0], np.ndarray):
                c = np.argmax(self.labels, axis=-1)
            else:
                c = self.labels
            x, y = rus.fit_sample(self.idxs.reshape(-1, 1), c[self.idxs])
            x = x.flatten()
            #y = y.flatten()
            np.random.shuffle(x)
            #print(len(x))
            #print(np.unique(y, return_counts=True))
            #np.random.shuffle(y)
            #for i in range(100):
            #    print(y[i], end="")
            #print()
            self.idxs = x
        elif self.shuffle:
            np.random.shuffle(self.idxs)

    def generator(self):
        """
        Generates pairs of data and optionally, cls. After all data is processed, the index is randomised
        :return: generator of (data[i], cls[i])
        """
        i = 0
        while i < len(self.idxs):
            idx = self.idxs[i]
            if self.labels is None:
                yield self.data[idx]
            else:
                yield (self.data[idx], self.labels[idx])
            i += 1
        self.on_epoch_end()

    def to_tfdataset(self):
        """
        Creates the tf.data.Dataset
        :return: A tf.data.Dataset that iterates through batches of (data, label) pairs
        """
        if isinstance(self.labels, list) or np.ndim(self.labels) <= 1:
            label_shape = None
        else:
            label_shape = self.labels[0].shape
        if self.labels is not None:
            ds = tf.data.Dataset.from_generator(self.generator,
                                                output_types=(self.data_dtype, self.labels_dtype),
                                                output_shapes=(self.data[0].shape, label_shape))
        else:
            ds = tf.data.Dataset.from_generator(self.generator,
                                                output_types=self.data_dtype,
                                                output_shapes=self.data[0].shape)
        """
        From docs:
        Performance can often be improved by setting num_parallel_calls so that map will use multiple threads to process elements. 
        If deterministic order isn't required, it can also improve performance to set deterministic=False.

        Note that the map function has to take a Tensor input
        """
        if self.map_fn is not None:
            ds = ds.map(lambda x, y: (self.map_fn(x), y), num_parallel_calls=tf.data.experimental.AUTOTUNE)
        if self.one_shot is False:
            ds = ds.repeat()
        ds = ds.batch(self.batch_size).prefetch(self.prefetch)
        return ds

    def tf1_compat_generator_error(self):
        # Get shapes from input data
        images = self.data
        onehots = self.labels
        shuffle_size = 1000
        img_size = images.shape
        img_size = (None, *img_size[1:])
        onehot_size = onehots.shape
        onehot_size = (None, onehot_size[1])
        images_tensor = tf.placeholder(tf.float32, shape=img_size)
        onehots_tensor = tf.placeholder(tf.float32, shape=onehot_size)
        # Create dataset
        dataset = tf.data.Dataset.from_tensor_slices((images_tensor, onehots_tensor))
        if self.map_fn is not None:
            dataset = dataset.map(lambda x, y: (self.map_fn(x), y), num_parallel_calls=tf.data.experimental.AUTOTUNE)
        if self.one_shot:
            dataset = dataset.repeat(1)
        else:
            dataset = dataset.shuffle(shuffle_size, reshuffle_each_iteration=True).repeat()
        dataset = dataset.batch(self.batch_size)
        dataset = dataset.prefetch(1)
        iterator = dataset.make_initializable_iterator()
        init_op = iterator.initializer
        next_val = iterator.get_next()
        with K.get_session().as_default() as sess:
            sess.run(init_op, feed_dict={images_tensor: images, onehots_tensor: onehots})
            while True:
                inputs, labels = sess.run(next_val)
                yield inputs, labels

    def tf1_compat_generator(self):
        if isinstance(self.labels, list) or np.ndim(self.labels) <= 1:
            label_shape = None
        else:
            label_shape = self.labels[0].shape
        if self.labels is not None:
            ds = tf.data.Dataset.from_generator(self.generator,
                                                output_types=(self.data_dtype, self.labels_dtype),
                                                output_shapes=(self.data[0].shape, label_shape))
        else:
            ds = tf.data.Dataset.from_generator(self.generator,
                                                output_types=self.data_dtype,
                                                output_shapes=self.data[0].shape)
        if self.map_fn is not None:
            ds = ds.map(lambda x, y: (self.map_fn(x), y), num_parallel_calls=tf.data.experimental.AUTOTUNE)
        if self.one_shot is False:
            ds = ds.repeat()
        ds = ds.batch(self.batch_size).prefetch(self.prefetch)
        iterator = ds.make_initializable_iterator()
        init_op = iterator.initializer
        next_val = iterator.get_next()
        with K.get_session().as_default() as sess:
            sess.run(init_op)
            while True:
                inputs, labels = sess.run(next_val)
                yield inputs, labels

    def create(self):
        if int(tf.__version__[0]) == 2:
            return self.to_tfdataset()
        else:
            return self.tf1_compat_generator()

    @staticmethod
    def map_fn_divide_255(t):
        t = tf.cast(t, tf.float32)
        return tf.divide(t, 255.0)

    @staticmethod
    def map_fn_divide_255_and_rotate_fn(k):
        def wrapper(t):
            t = tf.cast(t, tf.float32)
            t = tf.image.rot90(t, k)
            return tf.divide(t, 255.0)
        return wrapper




"""
Set of methods to make some common plots for the CNN outputs
"""
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.offsetbox import OffsetImage, AnnotationBbox
from scipy.spatial.distance import cdist
import scipy.ndimage as nd


def plot_embedding(X, y, num_classes, labels=None, title=None, indices=None, alpha=1.0, figsize=(8,8)):
    """
    Scatter plot of classes and their 2D embedding
    :param X: The 2D co-ordinates of each class value
    :param y: The classes
    :param num_classes: Total number of classes
    :param labels: Names of each class (if None, y values is used)
    :param title: Title of the plot
    :param indices: Extra label to append to each class point, e.g. the index of the image
    :param alpha: Transparency of the class label (0 - fully transparent, 1 - fully opaque)
    :param figsize: Size of the figure
    :return:
    """
    if labels is None:
        labels = ["{}".format(i) for i in range(num_classes)]
    # print(labels)
    # TODO: This can be simplified
    target_values = np.array(range(num_classes))
    colors = target_values / num_classes
    plt.figure(figsize=figsize)
    x_min, x_max = np.min(X, 0), np.max(X, 0)
    # X = (X - x_min) / (x_max - x_min)
    plt.axis([x_min[0], x_max[0], x_min[1], x_max[1]])
    points = [[] for k in range(len(target_values))]
    for i in range(X.shape[0]):
        for j in range(len(target_values)):
            if y[i] == target_values[j]:
                lbl = "{}".format(labels[y[i]])
                if indices is not None:
                    lbl += " {}".format(indices[i])
                plt.text(X[i, 0], X[i, 1], lbl,
                         color=plt.cm.tab20(colors[j]),
                         fontdict={'weight': 'bold', 'size': 9},alpha=alpha)
                points[j].append([X[i, 0], X[i, 1]])
                break
    plt.xticks([]), plt.yticks([])
    plt.tight_layout()
    if title is not None:
        plt.title(title)
    plt.gca().axis('off')


def plot_embedding_with_images(X, y, num_classes, images, face_colour="black", class_border=False, mask_level=(0.3, 0.7), scale_adj=1.0, figsize=None):
    """
    Scatter plot of classes and their 2D embedding. Instead of class name, the actual image is used.
    The dark parts of the images will be masked (used mainly for foraminifera images)
    :param X: The 2D co-ordinates of each class value
    :param y: The classes
    :param num_classes: Total number of classes
    :param images: The images corresponding to each class entry
    :param face_colour: Colour of the figure background
    :param class_border: Make a border with the colour of the class (not implemented)
    :param mask_level: Two values [0] is mask threshold, [1] is intensity above which everything is opaque
    :param scale_adj:
    :param figsize: Size of the figure. If None, it is scaled automatically (recommended)
    :return:
    """
    # Normalise the co-ordinates
    x_min, x_max = np.min(X, 0), np.max(X, 0)
    X = (X - x_min) / (x_max - x_min)
    # Calculate the scaling amount
    scale = images.shape[1] // 64 / scale_adj
    # Create the figure
    if figsize is None:
        figsize = (20*scale, 20*scale)
    plt.figure(figsize=figsize, dpi=50, facecolor=face_colour, edgecolor=face_colour)
    plt.gca().set_facecolor(face_colour)
    # Plot
    clrs = plt.cm.get_cmap('tab20')
    for j in range(X.shape[0]):
        img = np.squeeze(images[j,:,:,0])
        # Create a mask with a soft border
        # mask_level[0] is the threshold
        # mask_level[1] is the intensity value above which everything is fully opaque
        mask = img > mask_level[0]
        mask = nd.morphology.binary_fill_holes(mask)
        # if class_border:
        #     clr = clrs(y/num_classes)
        #     cmask = np.stack((mask*clr[0], mask*clr[1], mask*clr[2], np.ones(mask.shape)), axis=-1)
        th = img / mask_level[1]
        th[th > 1] = 1
        th[mask == True] = 1
        img = np.stack((img, img, img, th), axis=-1)
        ab = AnnotationBbox(OffsetImage(img), (0.03 + X[j, 0]*0.94, 0.03 + X[j, 1] * 0.94), xycoords="axes fraction", frameon=False)
        plt.gca().add_artist(ab)
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()


def plot_embedding_with_colour_images(X, images, face_colour="black", scale_adj=1.0, figsize=(10,10)):
    # Normalise the co-ordinates
    x_min, x_max = np.min(X, 0), np.max(X, 0)
    X = (X - x_min) / (x_max - x_min)
    # Calculate the scaling amount
    scale = images.shape[1] // 64 / scale_adj
    # Create the figure
    if figsize is None:
        figsize = (20*scale, 20*scale)
    plt.figure(figsize=figsize, dpi=50, facecolor=face_colour, edgecolor=face_colour)
    plt.gca().set_facecolor(face_colour)
    # Plot
    for j in range(X.shape[0]):
        img = images[j,:,:,:]
        ab = AnnotationBbox(OffsetImage(img), (0.03 + X[j, 0]*0.94, 0.03 + X[j, 1] * 0.94), xycoords="axes fraction", frameon=False)
        plt.gca().add_artist(ab)
    plt.xticks([])
    plt.yticks([])
    plt.tight_layout()


def plot_embedding_with_images_on_grid(X, y, num_classes, images, width=128, mask_level=0.7, scale_adj=1.0):
    """
    2D image scatter plots, but with the images in a grid.
    Requires lapjv which can be installed with 'pip install lapjv' BUT you need a C compiler installed as well
    :param X:
    :param y:
    :param num_classes:
    :param images:
    :param width:
    :param mask_level:
    :param scale_adj:
    :return:
    """
    grid_width = int(np.min((np.floor(np.sqrt(len(y))),20)))
    grid = np.dstack(np.meshgrid(np.linspace(0, 1, grid_width), np.linspace(0, 1, grid_width))).reshape(-1, 2)
    X = X[0:grid_width*grid_width,:]
    x_min, x_max = np.min(X, 0), np.max(X, 0)
    X = (X - x_min) / (x_max - x_min)
    cost_matrix = cdist(grid, X, "sqeuclidean").astype(np.float32)
    cost_matrix = cost_matrix * (100000 / cost_matrix.max())
    row_asses, col_asses, _ = lapjv(cost_matrix)
    grid_jv = grid[col_asses]
    plot_embedding_with_images(grid_jv, y, num_classes, images, face_colour="black", class_border=False, mask_level=mask_level, scale_adj=scale_adj)



"""
Plots of training and accuracy
"""
import numpy as np
from tensorflow.keras.callbacks import History
import matplotlib.pyplot as plt


def plot_loss_vs_epochs(history: History, figsize=(8, 4)):
    epochs = history.epoch
    loss = history.history['loss']
    val_loss = history.history['val_loss']
    plt.figure(figsize=figsize)
    plt.plot(epochs, loss)
    plt.plot(epochs, val_loss)
    plt.grid()
    plt.title("Loss")
    plt.legend(("Training", "Validation"))
    plt.xlim(left=0)
    plt.ylim(bottom=0)


def plot_accuracy_vs_epochs(history: History, metric='acc', figsize=(8, 4)):
    epochs = history.epoch
    acc = np.asarray(history.history[metric])
    val_acc = np.asarray(history.history['val_' + metric])
    plt.figure(figsize=figsize)
    plt.plot(epochs, acc * 100)
    plt.plot(epochs, val_acc * 100)
    plt.grid()
    plt.title("Accuracy")
    plt.legend(("Training", "Validation"))
    plt.xlim(left=0)
    plt.ylim(top=100)





"""
Performs inference on a directory of images organised by subdirectories, returning a pandas dataframe with the filename
and its classification
"""
import tensorflow as tf
from keras.utils import Sequence, OrderedEnqueuer
import numpy as np
import lxml.etree as ET
from tensorflow.python.platform import gfile
import keras.backend as K
import os
import pandas as pd
import multiprocessing
from pathlib import Path


class InferenceGenerator(Sequence):
    def __init__(self, source_dir, batch_size, img_size, img_type):
        self.source_dir = source_dir
        self.batch_size = batch_size
        self.img_size = img_size
        self.img_type = img_type
        filenames = DataSource.parse_directory(source_dir)
        self.filenames = [v for key, val in filenames.items() for v in val]

    def __data_generation(self, filenames):
        images = []
        for filename in filenames:
            image = DataSource.load_image(filename, self.img_size, self.img_type)
            image = DataSource.preprocess_image(image)
            images.append(image)
        return filenames, np.asarray(images)

    def __getitem__(self, index):
        batch_filenames = self.filenames[index * self.batch_size:(index + 1) * self.batch_size]
        filenames_and_images = self.__data_generation(batch_filenames)
        return filenames_and_images

    def __len__(self):
        return int(np.ceil(len(self.filenames) / self.batch_size))


def load_from_frozen(source: str,
                     input_tensor="input_1:0",
                     output_tensor="conv2d_23/Sigmoid:0",
                     session=None):
    if not os.path.exists(source):
        raise FileNotFoundError('The graph file was not found on the system.\nThe path was: ' + source)
    # Load network
    if session is None:
        session = K.get_session()
    with gfile.Open(source, 'rb') as f:
        graph_def = tf.GraphDef()
        graph_def.ParseFromString(f.read())
        session.graph.as_default()
        tf.import_graph_def(graph_def, name='')
    input = session.graph.get_tensor_by_name(input_tensor)
    output = session.graph.get_tensor_by_name(output_tensor)
    return session, input, output


def load_from_xml(filename, session=None):
    project = ET.parse(filename).getroot()

    protobuf = project.find('protobuf').text
    print("protobuf: " + protobuf)

    input = None
    output = None
    img_size = np.zeros(3, dtype=int)
    cls_labels = []

    list_xml = project.find('labels')
    for i, entry_xml  in enumerate(list_xml.iter('label')):
        code = entry_xml.find('code').text
        cls_labels.append(code)

    list_xml = project.find('inputs')
    for i, entry_xml in enumerate(list_xml.iter('input')):
        if i == 0:
            input_name = entry_xml.find('operation').text + ":0"
            img_size[0] = int(entry_xml.find('height').text)
            img_size[1] = int(entry_xml.find('width').text)
            img_size[2] = int(entry_xml.find('channels').text)

    list_xml = project.find('outputs')
    for i, entry_xml in enumerate(list_xml.iter('output')):
        if i == 0:
            output_name = entry_xml.find('operation').text + ":0"

    full_protobuf_path = os.path.join(os.path.dirname(filename), protobuf)
    session, input, output = load_from_frozen(full_protobuf_path, input_name, output_name)
    return session, input, output, img_size, cls_labels


def process(network_info, images_dir, output_dir, threshold=0.8):
    session, input_tensor, output_tensor, img_size, cls_labels = load_from_xml(network_info)
    print("Input tensor: {}".format(input_tensor))
    print("Output tensor: {}".format(output_tensor))
    print("Image size: {}".format(img_size))
    print("Classes: {}".format(cls_labels))
    print()
    print("Parsing source directory... (this can take some time)")
    if img_size[2] == 3:
        img_type = 'rgb'
    else:
        img_type = 'greyscale'
    gen = InferenceGenerator(images_dir, 64, img_size, img_type)

    print("Files: {}".format(len(gen.filenames)))
    print("Batches: {}".format(len(gen)))

    workers = np.min((multiprocessing.cpu_count(), 8))
    print("Workers: {}".format(workers))
    print()

    filenames = []
    cls_index = []
    cls_names = []
    score = []
    enq = OrderedEnqueuer(gen, use_multiprocessing=True)
    enq.start(workers=workers, max_queue_size=multiprocessing.cpu_count()*4)
    output_generator = enq.get()
    for i in range(len(gen)):
        print("\r{} / {}".format(i, len(gen)), end='')
        batch_filenames, batch_images = next(output_generator)
        result = session.run(output_tensor, feed_dict={input_tensor: batch_images})
        cls = np.argmax(result, axis=1)
        scr = np.max(result, axis=1)
        cls_name = [cls_labels[i] for i in cls]
        filenames.extend(batch_filenames)
        cls_index.extend(cls)
        cls_names.extend(cls_name)
        score.extend(scr)
    enq.stop()
    print()
    print("Done")
    print("See {} for results".format(output_dir))

    parents = [Path(f).parent.name for f in filenames]
    files = [Path(f).name for f in filenames]

    df = pd.DataFrame(data={'filename': filenames, 'parent': parents, 'file': files, 'class': cls_names, 'class_index': cls_index, 'score': score})
    os.makedirs(output_directory, exist_ok=True)
    df.to_csv(os.path.join(output_dir, "inference.csv"))


if __name__ == "__main__":
    network_info = r"D:\Development\Microfossil\PaperExperiments\Paper1CNNForMicrofossil\MD972138\Output\MD972138_base_cyclic_8_20200410-173126\model\network_info.xml"
    images_directory = r"D:\Datasets\Foraminifera\images_20200226_114546 Fifth with background"
    output_directory = r"D:\output"
    process(network_info, images_directory, output_directory)




import hashlib
import os
import numpy as np
from numpy.lib.format import open_memmap
import gc


class DatasetBase:
    def __init__(self, memmap_directory=None, overwrite_memmap=False, dtype=np.uint8):
        self.memmap_directory = memmap_directory
        self.overwrite_memmap = overwrite_memmap
        self.memmap_file = None
        self.hash_data = None
        self.data = None
        self.shape = None
        self.dtype = dtype

    def read_or_create_data(self, shape, dtype):
        self.shape = shape
        self.dtype = dtype

        if isinstance(shape, list):
            raise ValueError("Shape must be a tuple, e.g. (100,32,32,3) not a list, e.g. [100,32,32,3]")

        if self.memmap_directory:
            if self.hash_data is None:
                raise ValueError("Please set the hash_data (as a numpy array) to use memory mapping")
            self.memmap_file = os.path.join(self.memmap_directory, self.get_hash_id() + ".npy")
            if self.overwrite_memmap is False and os.path.exists(self.memmap_file):
                print("Existing data found at {}".format(self.memmap_file))
                self.data = open_memmap(self.memmap_file, mode='r+', dtype=self.dtype, shape=self.shape)
                # Check if not all zeros
                # If all zeros, usually indication of an error creating the memmap file previously, therefore recreate
                if np.count_nonzero(self.data[0]) > 0 and np.count_nonzero(self.data[-1]) > 0:
                    return True
                else:
                    self.data._mmap.close()
                    print("File was likely corrupted, recreating.")
            print("Creating memmap file at {}".format(self.memmap_file))
            os.makedirs(self.memmap_directory, exist_ok=True)
            self.data = open_memmap(self.memmap_file, mode='w+', dtype=self.dtype, shape=self.shape)
        else:
            self.data = np.zeros(self.shape, dtype=self.dtype)

        return False

    def get_hash_id(self):
        """
        Creates a 16 character hash id based on the hash data. This is used to create a unique
        filename for the numpy memmap array
        :return: 16 character hash id
        """
        if self.hash_data is None:
            raise ValueError("Please set the hash_data before getting the hash")
        else:
            return hashlib.sha256(repr(self.hash_data).encode('UTF-8')).hexdigest()[0:16]

    def release(self):
        if self.memmap_file is None:
            del self.data
        else:
            if os.path.exists(self.memmap_file):
                self.data._mmap.close()
                del self.data
                gc.collect()
                os.remove(self.memmap_file)
            self.memmap_file = None





import cv2
import numpy as np

class Morphology:
    def __init__(self):
        self.height = 0
        self.width = 0
        self.area = 0.0
        self.perimeter = 0.0
        self.convex_area = 0.0
        self.convex_perimeter = 0.0
        self.circle_area = 0.0
        self.circle_radius = 0.0
        self.major_axis_length = 0.0
        self.minor_axis_length = 0.0
        self.mean_diameter = 0.0
        self.eccentricity = 0.0
        self.angle = 0.0
        self.solidity = 0.0
        self.circularity = 0.0
        self.roundness = 0.0
        self.convex_perimeter_to_perimeter_ratio = 0.0
        self.convex_area_to_area_ratio = 0.0
        self.mean = 0.0
        self.stddev = 0.0
        self.stddev_invariant = 0.0
        self.skew = 0.0
        self.kurtosis = 0.0
        self.moment5 = 0.0
        self.moment6 = 0.0
        self.aspect_ratio = 0.0
        self.equivalent_diameter = 0.0
        self.perimeter_to_area_ratio = 0.0
        self.area_to_bounding_rectangle_area = 0.0
        self.equivalent_spherical_diameter = 0.0
        self.elongation = 0.0
        self.husmoment1 = 0.0
        self.husmoment2 = 0.0
        self.husmoment3 = 0.0
        self.husmoment4 = 0.0
        self.husmoment5 = 0.0
        self.husmoment6 = 0.0
        self.husmoment7 = 0.0

    def convert_to_mm(self, pixels_per_mm):
        pixels2_per_mm2 = pixels_per_mm * pixels_per_mm

        if pixels_per_mm == 0:
            return Morphology()

        self.mean_diameter /= pixels_per_mm
        self.area /= pixels2_per_mm2
        self.perimeter /= pixels_per_mm
        self.convex_area /= pixels2_per_mm2
        self.convex_perimeter /= pixels_per_mm
        self.circle_area /= pixels2_per_mm2
        self.circle_radius /= pixels_per_mm
        self.major_axis_length /= pixels_per_mm
        self.minor_axis_length /= pixels_per_mm

        return self



class MorphologyProcessor:
    @staticmethod
    def calculate_morphology(rgb, greyscale, contour):
        m = Morphology()

        # Exit if mask is non-existent or too small
        if contour is None or len(contour) < 5:
            return m

        contour = contour.astype(int)

        # Create binary image from contour
        binary = np.zeros(greyscale.shape, dtype=np.uint8)
        cv2.drawContours(binary, [contour], -1, 1, thickness=cv2.FILLED)
        binaryF = binary.astype(np.float32)

        m.area = cv2.contourArea(contour)

        if m.area < 25:
            return m

        m.width = greyscale.shape[1]
        m.height = greyscale.shape[0]
        m.mean_diameter = np.sqrt(m.area / np.pi) * 2

        # IMAGE INTENSITY MOMENTS
        # Mean(0) and standard deviation(1)
        mean, stddev = cv2.meanStdDev(greyscale, mask=binary)
        m.mean = mean[0][0]
        m.stddev = stddev[0][0]
        m.stddev_invariant = m.stddev / m.mean

        variance = np.power(m.stddev, 2)
        # num_pixels_in_mask = cv2.sumElems(binaryF)[0]
        # print(contour)
        # print(num_pixels_in_mask)
        num_pixels_in_mask = m.area

        greyscale_masked = cv2.multiply(binaryF, greyscale, dtype=cv2.CV_32FC1)

        # Mean
        m.mean = cv2.sumElems(greyscale_masked)[0] / num_pixels_in_mask

        # Subtract mean
        greyscale_masked = cv2.subtract(greyscale_masked, m.mean)
        greyscale_masked = cv2.multiply(greyscale_masked, binaryF, dtype=cv2.CV_32FC1)

        # Standard deviation
        temp = cv2.pow(greyscale_masked, 2)
        temp_sum = cv2.sumElems(temp)[0]
        m.stddev = np.sqrt(temp_sum / num_pixels_in_mask)

        # Skew
        temp = cv2.pow(greyscale_masked, 3)
        temp_sum = cv2.sumElems(temp)[0]
        m.skew = temp_sum / np.power(m.stddev, 3) / num_pixels_in_mask

        # Kurtosis
        temp = cv2.pow(greyscale_masked, 4)
        temp_sum = cv2.sumElems(temp)[0]
        m.kurtosis = temp_sum / np.power(m.stddev, 4) / num_pixels_in_mask

        # 5th moment
        temp = cv2.pow(greyscale_masked, 5)
        temp_sum = cv2.sumElems(temp)[0]
        m.moment5 = temp_sum / np.power(m.stddev, 5) / num_pixels_in_mask

        # 6th moment
        temp = cv2.pow(greyscale_masked, 6)
        temp_sum = cv2.sumElems(temp)[0]
        m.moment6 = temp_sum / np.power(m.stddev, 6) / num_pixels_in_mask

        # MASK DIMENSIONS
        # Area and perimeter
        m.perimeter = cv2.arcLength(contour, True)

        # Convex area and perimeter
        convex_contour = cv2.convexHull(contour)
        # print(hull)
        # convex_contour = np.array([contour[idx] for idx in hull])

        m.convex_area = cv2.contourArea(convex_contour)
        m.convex_perimeter = cv2.arcLength(convex_contour, True)
        m.convex_perimeter_to_perimeter_ratio = m.convex_perimeter / m.perimeter
        m.convex_area_to_area_ratio = m.convex_area / m.area

        # ELLIPSE AND CIRCLE
        # Enclosing circle
        (x, y), radius = cv2.minEnclosingCircle(contour)
        m.circle_radius = radius
        m.circle_area = np.pi * radius * radius

        # Approximating ellipse
        if len(contour) >= 5:
            (center, (major_axis_length, minor_axis_length), angle) = cv2.fitEllipse(contour)
            m.major_axis_length = major_axis_length
            m.minor_axis_length = minor_axis_length
            m.eccentricity = np.sqrt(1 - np.power(m.minor_axis_length / m.major_axis_length, 2))
            m.angle = angle
            m.roundness = 4.0 * m.area / np.pi / np.power(m.major_axis_length, 2)
            m.elongation = m.major_axis_length / m.minor_axis_length

        # COMMON MEASUREMENTS
        m.solidity = m.area / m.convex_area
        m.circularity = 4.0 * np.pi * m.area / np.power(m.perimeter, 2)
        m.perimeter_to_area_ratio = m.perimeter / m.area
        m.equivalent_diameter = np.sqrt(4.0 * m.area / np.pi)
        m.equivalent_spherical_diameter = 2.0 * np.sqrt(m.area / np.pi)

        bounding_rectangle = cv2.boundingRect(contour)
        m.aspect_ratio = bounding_rectangle[2] / bounding_rectangle[3]
        m.area_to_bounding_rectangle_area = m.area / (bounding_rectangle[2] * bounding_rectangle[3])

        hu_moments = cv2.HuMoments(cv2.moments(contour)).flatten()
        m.husmoment1 = hu_moments[0]
        m.husmoment2 = hu_moments[1]
        m.husmoment3 = hu_moments[2]
        m.husmoment4 = hu_moments[3]
        m.husmoment5 = hu_moments[4]
        m.husmoment6 = hu_moments[5]
        m.husmoment7 = hu_moments[6]

        return m


import tensorflow as tf
import numpy as np


class CyclicSlice4(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super(CyclicSlice4, self).__init__(**kwargs)

    def call(self, input):
        F = slice_4(input)
        return F

class CyclicGainSlice12(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super(CyclicGainSlice12, self).__init__(**kwargs)

    def call(self, input):
        F = slice_gain_12(input)
        return F


class CyclicRoll4(tf.keras.layers.Layer):
    def __init__(self, **kwargs):
        super(CyclicRoll4, self).__init__(**kwargs)

    def call(self, input):
        return roll_4(input)


class CyclicPool4(tf.keras.layers.Layer):
    def __init__(self, pool_op, **kwargs):
        super(CyclicPool4, self).__init__(**kwargs)
        self.pool_op = pool_op

    def call(self, input):
        return pool_4(input, self.pool_op)


class CyclicDensePool4(tf.keras.layers.Layer):
    def __init__(self, pool_op):
        super(CyclicDensePool4, self).__init__()
        self.pool_op = pool_op

    def call(self, input):
        return dense_pool_4(input, self.pool_op)


class CyclicDensePoolN(tf.keras.layers.Layer):
    def __init__(self, pool_op, split_count=12):
        super(CyclicDensePoolN, self).__init__()
        self.pool_op = pool_op
        self.n = split_count

    def call(self, input):
        return dense_pool_n(input, self.pool_op, self.n)

def gain_3(X):
    Y = []
    Y.append(X)
    Y.append(tf.multiply(X, 0.5))
    Y.append(tf.multiply(X, 2))
    return Y

def rotate_4(X):
    Y = []
    Y.append(X)
    Y.append(tf.reverse(tf.transpose(X, [0, 2, 1, 3]), [1]))
    Y.append(tf.reverse(X, [1, 2]))
    Y.append(tf.reverse(tf.transpose(X, [0, 2, 1, 3]), [2]))
    return Y

def unrotate_4(X):
    Y = []
    Y.append(X)
    Y.append(tf.reverse(tf.transpose(X, [0, 2, 1, 3]), [2]))
    Y.append(tf.reverse(X, [1, 2]))
    Y.append(tf.reverse(tf.transpose(X, [0, 2, 1, 3]), [1]))
    return Y

def reorder_4(X, order):
    Y = tf.split(X, 4)
    Z = [Y[i] for i in order]
    return tf.concat(Z, 0)


def slice_4(X):
    Y = rotate_4(X)
    return tf.concat(Y, 0)

def slice_gain_12(X):
    Y = gain_3(X)
    Y = tf.concat(Y, 0)
    Y = rotate_4(Y)
    return tf.concat(Y, 0)

# Also try role that is a bit different
def roll_4(X):
    Y = unrotate_4(X)
    Z = []
    Z.append(X)
    for i in range(1, 4):
        Z.append(reorder_4(Y[i], np.roll(range(4), shift=-i)))
    return tf.concat(Z, 3)


def stack_4(X):
    Y = tf.split(X, 4)
    Z = []
    Z.append(Y[0])
    Z.append(tf.reverse(tf.transpose(Y[1], [0, 2, 1, 3]), [2]))
    Z.append(tf.reverse(Y[2], [1, 2]))
    Z.append(tf.reverse(tf.transpose(Y[3], [0, 2, 1, 3]), [1]))
    return tf.concat(3)


def pool_4(X, pool_op):
    Y = tf.split(X, 4)
    Z = []
    Z.append(Y[0])
    Z.append(tf.reverse(tf.transpose(Y[1], [0, 2, 1, 3]), [2]))
    Z.append(tf.reverse(Y[2], [1, 2]))
    Z.append(tf.reverse(tf.transpose(Y[3], [0, 2, 1, 3]), [1]))
    W = tf.stack(Z, 4)
    return pool_op(W, (4))


def dense_pool_4(X, pool_op):
    Y = tf.split(X, 4)
    W = tf.stack(Y, 2)
    return pool_op(W, 2)

def dense_pool_n(X, pool_op, n):
    Y = tf.split(X, n)
    W = tf.stack(Y, 2)
    return pool_op(W, 2)



import numpy as np
from PIL import Image, ImageSequence
from skimage.transform import resize
from skimage.color import rgb2gray
from sklearn.model_selection import train_test_split
import matplotlib.pyplot as plt
import glob
import os
import pandas as pd


class MaskDataSource:

    def __init__(self):
        self.data_df = None
        self.train_df = None
        self.test_df = None
        self.train_images = None
        self.train_masks = None
        self.test_images = None
        self.test_masks = None

    def load_images(self,
                    img_size,
                    rescale_params=(255, 0, 1),
                    color_mode='rgb',
                    split=0.25,
                    flatten=True,
                    seed=None):
        filenames = self.data_df['filenames']
        images = []
        masks = []
        # Load each image
        print("Loading images...")
        print()
        for idx, filename in enumerate(filenames):
            if idx % 100 == 0:
                print(idx)
            im = Image.open(filename)

            for i, page in enumerate(ImageSequence.Iterator(im)):
                # Colour image
                if i == 0:
                    page = np.array(page)
                    if color_mode == 'grayscale':
                        page = rgb2gray(page)
                    else:
                        page = page / 255
                    page = resize(page, img_size, order=1)
                    images.append(page)
                if i == 4:
                    page = np.array(page)
                    page = rgb2gray(page)
                    page = resize(page, img_size, order=1)
                    page = page > 0.5
                    page = page.astype(int)
                    if flatten:
                        encoded = np.zeros((img_size[0], img_size[1], 2))
                        for j in range(2):
                            temp = page == 0
                            temp = temp.astype(int)
                            encoded[:, :, j] = temp
                        encoded = np.reshape(encoded, (img_size[0] * img_size[1], 2))
                        masks.append(encoded)
                    else:
                        page = page[:,:,np.newaxis]
                        masks.append(page)
                if i == 5:
                    break
        images = np.asarray(images)
        if images.ndim == 3:
            images = images[:, :, :, np.newaxis]
        masks = np.asarray(masks)
        self.split(images, masks, split, seed)

    def split(self, images, masks, split=0.25, seed=None):
        self.train_images, self.test_images, \
        self.train_masks, self.test_masks, \
        self.train_df, self.test_df = \
            train_test_split(images,
                             masks,
                             self.data_df,
                             test_size=split,
                             random_state=seed)

    def set_directory_source(self,
                             directory):
        files = glob.glob(directory + "\\*.tiff")
        files.extend(glob.glob(directory + "\\*.tif"))
        files = sorted(files)

        filenames = []
        short_filenames = []
        for file in files:
            filenames.append(file)
            short_filenames.append(
                os.path.basename(os.path.dirname(file)) + os.path.sep + os.path.basename(file))

        df = {"filenames": filenames, "short_filenames": short_filenames}
        self.data_df = pd.DataFrame(df)
        print("{} files found".format(len(filenames)))

        #     if img_type == 'rgb':
        #         im = Image.open(filename)
        #     else:
        #         im = Image.open(filename).convert('L')
        #     im = np.asarray(im, dtype=np.float)
        #     im = resize(im, img_size, order=1)
        #     # Rescale according to params
        #     # e.g. to take a [0,255] range image to [-1,1] params would be 255,-0.5,2
        #     # I' = (I / 255 - 0.5) * 2
        #     im = np.divide(im, rescale_params[0])
        #     im = np.subtract(im, rescale_params[1])
        #     im = np.multiply(im, rescale_params[2])
        #     # Convert to float
        #     im = im.astype(np.float32)
        #     images.append(im)
        #     idx += 1
        #     if idx % 100 == 0:
        #         print("\r{} of {} processed".format(idx, len(filenames)), end='')
        # # Convert all to a numpy array
        # images = np.asarray(images)
        # if images.ndim == 3:
        #     images = images[:, :, :, np.newaxis]
        # # Split into test and training sets
        # self.split(images, split, seed)

# d = DataSource()
# d.set_directory_source(r"C:\Users\rossm\Documents\Data\Foraminifera\Collated\MD042722_200-201_()_18")
# d.load_images((128,128))



import numpy as np
import itertools
import matplotlib.pyplot as plt
import matplotlib.patches as pch
import matplotlib.lines as lines
from sklearn.metrics import confusion_matrix, precision_recall_fscore_support, accuracy_score
from matplotlib.collections import PatchCollection


def plot_confusion_matrix(y_true,
                          y_pred,
                          cls_labels,
                          normalise=True,
                          title='Confusion matrix',
                          cmap=plt.cm.Blues,
                          figsize=None,
                          style='checker',
                          show=False):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    # If figsize is None, estimate the plot size
    if figsize is None:
        figsize = (len(cls_labels) / 3, len(cls_labels) / 3)
    # Calculate precision, recall, etc
    p, r, f1, s = precision_recall_fscore_support(y_true, y_pred, labels=range(len(cls_labels)))
    # Calculate confusion matrix
    cm = confusion_matrix(y_true=y_true,
                          y_pred=y_pred,
                          labels=range(len(cls_labels)))
    # Normalise the values
    if normalise:
        count = cm.sum(axis=1)[:, np.newaxis]
        count[count == 0] = 1
        cm = cm.astype('float') / count
        cm = np.nan_to_num(cm)
        cm = np.round(cm * 100).astype(int)
    # Plot confusion matrix
    plt.figure(facecolor="white", figsize=figsize)
    plt.imshow(cm, interpolation='nearest', cmap=cmap)
    plt.title(title)
    # Add axes cls
    tick_marks = np.arange(len(cls_labels))
    plt.xticks(tick_marks, cls_labels, rotation=90)
    plt.yticks(tick_marks, cls_labels)
    # Add grid
    if style == 'grid':
        ax = plt.gca()
        minor_tick_marks = tick_marks[:-1] + 0.5
        ax.set_xticks(minor_tick_marks, minor=True)
        ax.set_yticks(minor_tick_marks, minor=True)
        plt.grid(which='minor')
    # Add percentages in boxes
    thresh = cm.max() / 2.
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        if cm[i, j] != 0 or style == 'checker':
            plt.text(j, i + 0.25, cm[i, j],
                     horizontalalignment="center",
                     color="white" if cm[i, j] > thresh else "black")
    plt.ylabel('True label')
    plt.xlabel('Predicted label')
    plt.tight_layout()
    if show is True:
        plt.show()


def get_text_width(txt):
    f = plt.figure()
    r = f.canvas.get_renderer()
    t = plt.text(0.5, 0.5, txt)
    plt.tight_layout()
    bb = t.get_window_extent(renderer=r)
    width = bb.width
    plt.close('all')
    return width


def plot_confusion_accuracy_matrix(y_true,
                                   y_pred,
                                   cls_labels,
                                   normalise=True,
                                   title='Confusion matrix',
                                   cmap=plt.cm.Blues,
                                   figsize=None,
                                   style='grid5',
                                   show=False):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    mult = 3
    # If figsize is None, estimate the plot size
    max_word = cls_labels[np.argmax([len(f"{lab}") for lab in cls_labels])]
    txt_width = get_text_width(max_word) / 40
    # print(txt_width)
    sz = len(cls_labels) / mult + txt_width + 1.5
    sz2 = len(cls_labels) / mult + 1.5
    if figsize is None:
        figsize = (sz, sz)
    # Calculate precision, recall, etc
    p, r, f1, s = precision_recall_fscore_support(y_true, y_pred, labels=range(len(cls_labels)))
    # Calculate confusion matrix
    cm = confusion_matrix(y_true=y_true,
                          y_pred=y_pred,
                          labels=range(len(cls_labels)))
    # Normalise the values
    if normalise:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        cm = np.nan_to_num(cm)
        cm = np.round(cm * 100).astype(int)
    # Create combined plots
    f, ax = plt.subplots(2, 2,
                         gridspec_kw={'width_ratios': [sz2 - 1.5, 1],
                                      'height_ratios': [1, sz2 - 1.5],
                                      'wspace': 0,
                                      'hspace': 0},
                         figsize=figsize)
    # Add counts to class cls
    cls_labels = ['{} ({})'.format(cls_labels[i], s[i]) for i in range(len(cls_labels))]
    thresh = cm.max() / 2.
    # Axes
    ax_cm = ax[1, 0]
    ax_right = ax[1, 1]
    ax_top = ax[0, 0]
    ax_unused = ax[0, 1]
    ax_unused.axis('off')
    # Axes cls
    tick_marks = np.arange(len(cls_labels))
    ax_cm.set_xticks(tick_marks)
    ax_cm.set_xticklabels(cls_labels, rotation=90)
    ax_cm.set_yticks(tick_marks)
    ax_cm.set_yticklabels(cls_labels)
    # ax_cm.set_xticks(tick_marks)
    # ax_cm.set_xticklabels(cls_labels)
    ax_cm.set_xlim(-0.5, len(cls_labels) - 0.5)
    ax_cm.set_ylim(-0.5, len(cls_labels) - 0.5)
    ax_cm.invert_yaxis()
    # Bar plots
    remove_frame(ax_top)
    ax_top.bar(tick_marks, p, width=0.8, color=cmap(p), edgecolor=(0, 0, 0, 0.6))
    ax_top.set_xlim((-0.5, len(tick_marks) - 0.5))
    for i, v in enumerate(p):
        if np.mean(cmap(v)[:-1]) < 0.5:
            clr = 'white'
        else:
            clr = 'black'
        ax_top.text(i, 0.15, '{:.1f}'.format(v * 100), color=clr, ha='center', rotation=90, alpha=0.7)
    remove_frame(ax_right)
    ax_right.barh(tick_marks, r, height=0.8, color=cmap(r), edgecolor=(0, 0, 0, 0.6))
    ax_right.set_ylim((-0.5, len(tick_marks) - 0.5))
    for i, v in enumerate(r):
        if np.mean(cmap(v)[:-1]) < 0.5:
            clr = 'white'
        else:
            clr = 'black'
        ax_right.text(0.05, i, '{:.1f}'.format(v * 100), color=clr, va='center', alpha=0.7)
    ax_right.invert_yaxis()
    # Confusion matrix
    patches = []
    colors = []
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        patches.append(pch.Rectangle((j - 0.5, i - 0.5), 1, 1))
        colors.append(cm[i, j] / 100)
        # if cm[i, j] != 0 or style == 'checker':
        ax_cm.text(j, i + 0.25, cm[i, j],
                   horizontalalignment="center",
                   color="white" if cm[i, j] > thresh else "black")
    patcol = PatchCollection(patches, alpha=1, cmap=cmap)
    patcol.set_array(np.array(colors))
    ax_cm.add_collection(patcol)
    # Grid
    if style == 'grid':
        for i in range(len(cls_labels) - 1):
            line1 = lines.Line2D([i + 0.5, i + 0.5], [-0.5, len(cls_labels) + 0.5], color=(0, 0, 0, 0.05))
            line2 = lines.Line2D([-0.5, len(cls_labels) + 0.5], [i + 0.5, i + 0.5], color=(0, 0, 0, 0.05))
            ax_cm.add_line(line1)
            ax_cm.add_line(line2)
    if style == 'grid5':
        for i in range(len(cls_labels) - 1):
            if i % 5 == 4:
                line1 = lines.Line2D([i + 0.5, i + 0.5], [-0.5, len(cls_labels) + 0.5], color=(0, 0, 0, 0.2))
                line2 = lines.Line2D([-0.5, len(cls_labels) + 0.5], [i + 0.5, i + 0.5], color=(0, 0, 0, 0.2))
                ax_cm.add_line(line1)
                ax_cm.add_line(line2)
    # Labels
    ax_cm.set_ylabel('True label')
    ax_cm.set_xlabel('Predicted label')
    ax_right.set_ylabel('Recall {:.1f}%'.format(np.mean(r) * 100))
    ax_right.yaxis.set_label_position('right')
    ax_top.set_xlabel('Precision {:.1f}%'.format(np.mean(p) * 100))
    ax_top.xaxis.set_label_position('top')
    ax_top.set_title('Overall accuracy {:.1f}%'.format(accuracy_score(y_true, y_pred) * 100))
    ax_cm.set_zorder(100)
    plt.tight_layout()
    if show is True:
        plt.show()


def remove_frame(ax):
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.get_xaxis().set_ticks([])
    ax.get_yaxis().set_ticks([])


def plot_comparison_matrix(y_true,
                           y_pred,
                           true_cls_labels,
                           pred_cls_labels,
                           normalise=True,
                           title='Comparison matrix',
                           cmap=plt.cm.Blues,
                           figsize=None,
                           style='grid5',
                           show=False
                           ):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    # If figsize is None, estimate the plot size
    if figsize is None:
        figsize = (len(pred_cls_labels) / 2.75 + 2, len(true_cls_labels) / 2.75 + 2)

    # Calculate confusion matrix
    max_labels = np.max((len(true_cls_labels), len(pred_cls_labels)))
    cm = confusion_matrix(y_true=y_true,
                          y_pred=y_pred,
                          labels=range(max_labels))
    cm = cm[0:len(true_cls_labels), 0:len(pred_cls_labels)]
    support = np.sum(cm, axis=1)
    # Normalise the values
    if normalise:
        cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]
        cm = np.round(cm * 100).astype(int)
    # Create combined plots
    # f, ax = plt.subplots(2, 2,
    #                      gridspec_kw={'width_ratios': [6, 1],
    #                                   'height_ratios': [1, 6],
    #                                   'wspace': 0,
    #                                   'hspace': 0},
    #                      figsize=figsize)
    # Add counts to class cls

    true_cls_labels = ['{} ({})'.format(true_cls_labels[i], support[i]) for i in range(len(true_cls_labels))]
    thresh = cm.max() / 2.

    fig, ax = plt.subplots(1, 1, figsize=figsize)
    # Axes cls
    true_tick_marks = np.arange(len(true_cls_labels))
    pred_tick_marks = np.arange(len(pred_cls_labels))

    ax.set_yticks(true_tick_marks)
    ax.set_yticklabels(true_cls_labels)
    ax.set_ylim(-0.5, len(true_cls_labels) - 0.5)

    ax.set_xticks(pred_tick_marks)
    ax.set_xticklabels(pred_cls_labels, rotation=90)
    ax.set_xlim(-0.5, len(pred_cls_labels) - 0.5)

    ax.invert_yaxis()

    # Confusion matrix
    patches = []
    colors = []
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        patches.append(pch.Rectangle((j - 0.5, i - 0.5), 1, 1))
        colors.append(cm[i, j] / 100)
        # if cm[i, j] != 0 or style == 'checker':
        ax.text(j, i + 0.25, cm[i, j],
                horizontalalignment="center",
                color="white" if cm[i, j] > thresh else "black")
    patcol = PatchCollection(patches, alpha=1, cmap=cmap)
    patcol.set_array(np.array(colors))
    ax.add_collection(patcol)

    # Grid
    if style == 'grid':
        for i in range(len(pred_cls_labels) - 1):
            line1 = lines.Line2D([i + 0.5, i + 0.5], [-0.5, len(true_cls_labels) + 0.5], color=(0, 0, 0, 0.05))
            ax.add_line(line1)
        for i in range(len(true_cls_labels) - 1):
            line2 = lines.Line2D([-0.5, len(pred_cls_labels) + 0.5], [i + 0.5, i + 0.5], color=(0, 0, 0, 0.05))
            ax.add_line(line2)

    if style == 'grid5':
        for i in range(len(pred_cls_labels) - 1):
            if i % 5 == 4:
                line1 = lines.Line2D([i + 0.5, i + 0.5], [-0.5, len(true_cls_labels) + 0.5], color=(0, 0, 0, 0.2))
                line2 = lines.Line2D([-0.5, len(true_cls_labels) + 0.5], [i + 0.5, i + 0.5], color=(0, 0, 0, 0.2))
                ax.add_line(line1)
                ax.add_line(line2)
        for i in range(len(true_cls_labels) - 1):
            if i % 5 == 4:
                line1 = lines.Line2D([i + 0.5, i + 0.5], [-0.5, len(pred_cls_labels) + 0.5], color=(0, 0, 0, 0.2))
                line2 = lines.Line2D([-0.5, len(pred_cls_labels) + 0.5], [i + 0.5, i + 0.5], color=(0, 0, 0, 0.2))
                ax.add_line(line1)
                ax.add_line(line2)
    # Labels
    ax.set_ylabel('True cls')
    ax.set_xlabel('Predicted cls')
    plt.tight_layout()
    if show is True:
        plt.show()

#
# if __name__ == "__main__":
#     import string
#     import random
#
#     def get_random_string(length):
#         # choose from all lowercase letter
#         letters = string.ascii_lowercase
#         result_str = ''.join(random.choice(letters) for i in range(length))
#         # print("Random string of length", length, "is:", result_str)
#         return result_str
#
#     offset = 0
#     for i in [3,4,30,50]:
#         y_true = np.arange(i+offset)
#         y_pred = np.arange(i+offset)
#         cls_labels = [get_random_string(np.random.randint(3,30)) for j in range(i+offset)]
#         cls_labels[0] = "DIAT-Chaetoceros"
#         cls_labels[1] = "DIAT-Other"
#         cls_labels[2] = "DIAT-Pseudo-nitzschia"
#         # cls_labels[0] = "DIAT-"
#         # cls_labels[1] = "DIAT-"
#         # cls_labels[2] = "DIAT-"
#         if len(cls_labels) > 3:
#             cls_labels[3] = "DIAT-Pseudo-nitzschia"
#         print(cls_labels)
#         plot_confusion_accuracy_matrix(y_true,
#                                        y_pred,
#                                        cls_labels,
#                                        normalise=True,
#                                        title='Confusion matrix',
#                                        cmap=plt.cm.Blues,
#                                        figsize=None,
#                                        style='grid5',
#                                        show=False)
#         plt.show()



import tensorflow as tf
import numpy as np
import tensorflow_addons as tfa


def aug_all_fn(rotation=(0, 360),
               gain=(0.8, 1.0, 1.2),
               gamma=(0.5, 1.0, 2),
               zoom=(0.9, 1.0, 1.1),
               gaussian_noise=None,
               bias=None,
               random_crop=None,
               divide=255):
    def wrapper(im_x):
        im_x = tf.cast(im_x, tf.float32)
        if divide is not None:
            im_x = tf.divide(im_x, tf.constant(divide, dtype=tf.float32))
        im_x = aug_rotation(im_x, rotation)
        im_x = aug_zoom(im_x, zoom)
        im_x = aug_gain_gamma(im_x, gain, gamma)
        im_x = aug_gaussian_noise(im_x, gaussian_noise)
        im_x = aug_bias(im_x, bias)
        im_x = aug_random_crop(im_x, random_crop)
        return im_x
    return wrapper


def aug_rotation(im_x, rotation=(0, 360)):
    if rotation is not None:
        if len(rotation) > 2:
            rotation_factor = tf.random.shuffle(tf.constant(rotation))[0]
        elif len(rotation) == 2:
            rotation_factor = tf.random.uniform([], rotation[0] / 180 * np.pi, rotation[1] / 180 * np.pi)
        else:
            raise ValueError("Rotation needs at least 2 values")
        im_x = tfa.image.rotate(im_x, rotation_factor, interpolation='bilinear')
    return im_x


def aug_zoom(im_x, zoom=(0.9, 1.0, 1.1)):
    if zoom is not None:
        if len(zoom) > 2:
            zoom_value = tf.random.shuffle(tf.constant(zoom))[0]
        elif len(zoom) == 2:
            zoom_value = tf.random.uniform([], zoom[0], zoom[1])
        else:
            raise ValueError("Zoom needs at least 2 values")
        zoom_start_factor = tf.divide(tf.subtract(1.0, zoom_value), 2)
        zoom_end_factor = tf.subtract(1.0, zoom_start_factor)
        zoom_factor = [[zoom_start_factor, zoom_start_factor, zoom_end_factor, zoom_end_factor]]
        im_x = tf.image.crop_and_resize([im_x],
                                        boxes=zoom_factor,
                                        box_indices=tf.constant([0]),
                                        crop_size=tf.shape(im_x)[0:2],
                                        method='bilinear',
                                        extrapolation_value=0)[0]
    return im_x


def aug_gain_gamma(im_x, gain=(0.8, 1.0, 1.2), gamma=(0.5, 1.0, 2)):
    # Gain
    if gain is not None:
        if len(gain) > 2:
            gain_factor = tf.random.shuffle(tf.constant(gain))[0]
        elif len(gain) == 2:
            gain_factor = tf.random.uniform([], gain[0], gain[1])
        else:
            raise ValueError("Gain needs at least 2 values")
    else:
        gain_factor = tf.constant(1.0)
    # Gamma
    if gamma is not None:
        if len(gamma) > 2:
            gamma_factor = tf.random.shuffle(tf.constant(gamma))[0]
        elif len(gamma) == 2:
            gamma_factor = tf.random.uniform([], gamma[0], gamma[1])
        else:
            raise ValueError("Gamma needs at least 2 values")
    else:
        gamma_factor = tf.constant(1.0)
    if gamma is not None or gain is not None:
        im_x = tf.image.adjust_gamma(im_x, gamma_factor, gain_factor)
    return im_x


def aug_gaussian_noise(im_x, gaussian_noise=None):
    shape = tf.shape(im_x)
    if gaussian_noise is not None:
        if len(gaussian_noise) > 2:
            gaussian_noise_factor = tf.random.shuffle(tf.constant(gaussian_noise))[0]
        elif len(gaussian_noise) == 2:
            gaussian_noise_factor = tf.random.uniform([], gaussian_noise[0], gaussian_noise[1])
        else:
            raise ValueError("Noise needs at least 2 values")
        noise_tensor = tf.random.normal(shape,
                                        mean=0.0,
                                        stddev=gaussian_noise_factor,
                                        dtype=tf.float32,
                                        seed=None,
                                        name=None)
        im_x = tf.add(im_x, noise_tensor)
    return im_x


def aug_bias(im_x, bias=(-0.5, 0.5)):
    if bias is not None:
        if len(bias) > 2:
            bias_factor = tf.random.shuffle(tf.constant(bias))[0]
        elif len(bias) == 2:
            bias_factor = tf.random.uniform([], bias[0], bias[1])
        else:
            raise ValueError("Offset needs at least 2 values")
        im_x = tf.add(im_x, bias_factor)
    return im_x


def aug_random_crop(im_x, crop_size):
    if crop_size is not None:
        im_x = tf.image.random_crop(im_x, crop_size)
    return im_x



import time
import tensorflow as tf
import numpy as np

from miso.data.tf_generator import TFGenerator
from miso.data.training_dataset import TrainingDataset
from miso.models.factory import generate_tl, combine_head_and_tail
from miso.training.adaptive_learning_rate import AdaptiveLearningRateScheduler
from miso.training.helpers import predict_in_batches
from miso.training.parameters import MisoParameters
from miso.training.augmentation import aug_all_fn


def train_transfer_learning(tp: MisoParameters, ds: TrainingDataset):

    print("-" * 80)
    print("Transfer learning network training")
    start = time.time()

    # Generate head and tail models separately
    model_head, model_tail = generate_tl(tp)
    # model_head.summary()

    # Calculate vectors using head model
    print("- calculating vectors")
    t = time.time()
    gen = ds.images.create_generator(
        tp.training.batch_size, shuffle=False, one_shot=True
    )
    vectors = model_head.predict(gen.create())
    print(f"! {time.time() - t}s elapsed, ({len(vectors)}/{len(ds.images.data)} vectors)")

    # Calculate augmented vectors
    train_idx = ds.train_idx
    cls_onehot = ds.cls_onehot
    if tp.training.use_augmentation is True and tp.training.transfer_learning_augmentation_factor > 0:
        print("- calculating augmentated vectors")
        t = time.time()
        if tp.augmentation.rotation is True:
            tp.augmentation.rotation = [0, 360]
        elif tp.augmentation.rotation is False:
            tp.augmentation.rotation = None

        augment_fn = aug_all_fn(
            rotation=tp.augmentation.rotation,
            gain=tp.augmentation.gain,
            gamma=tp.augmentation.gamma,
            zoom=tp.augmentation.zoom,
            gaussian_noise=tp.augmentation.gaussian_noise,
            bias=tp.augmentation.bias,
            random_crop=tp.augmentation.random_crop,
            divide=255,
        )

        aug_gen = ds.train_generator(
            batch_size=tp.training.batch_size,
            shuffle=False,
            one_shot=True,
            map_fn=augment_fn,
            undersample=tp.training.use_class_undersampling,
        )

        for aug_i in range(tp.training.transfer_learning_augmentation_factor):
            print(f"- {aug_i + 1}/{tp.training.transfer_learning_augmentation_factor}")
            aug_vectors = model_head.predict(aug_gen.create())
            len_vectors = len(vectors)
            train_idx = np.concatenate((train_idx, np.arange(len(aug_vectors)) + len_vectors))
            vectors = np.concatenate((vectors, aug_vectors))
            cls_onehot = np.concatenate((cls_onehot, ds.cls_onehot[ds.train_idx]))
        print(f"! {time.time() - t}s elapsed, ({len(vectors)}/{len(ds.images.data)} normal + augmented vectors)")

    print("-" * 80)
    print("Training")

    # Learning rate scheduler
    alr_cb = AdaptiveLearningRateScheduler(nb_epochs=tp.training.alr_epochs,
                                           nb_drops=tp.training.alr_drops,
                                           verbose=1)

    # Training generator
    train_gen = TFGenerator(vectors,
                            cls_onehot,
                            train_idx,
                            tp.training.batch_size,
                            shuffle=True,
                            one_shot=False,
                            undersample=tp.training.use_class_undersampling)

    # Validation generator
    if tp.dataset.val_split > 0:
        val_gen = TFGenerator(
            vectors,
            ds.cls_onehot,
            ds.test_idx,
            tp.training.batch_size,
            shuffle=False,
            one_shot=True,
        )
        val_data = val_gen.create()
        val_steps = len(val_gen)
    else:
        val_data = None
        val_steps = None

    # Class weights (only if over sampling is not used)
    if tp.training.use_class_weights is True and tp.training.use_class_undersampling is False:
        class_weights = ds.class_weights
        print("- class weights: {}".format(class_weights))
        class_weights = dict(enumerate(class_weights))
    else:
        class_weights = None
    if tp.training.use_class_undersampling:
        print("- class balancing using random under sampling")

    # Train
    history = model_tail.fit(train_gen.create(),
                             steps_per_epoch=len(train_gen),
                             validation_data=val_data,
                             validation_steps=val_steps,
                             epochs=tp.training.max_epochs,
                             verbose=0,
                             shuffle=False,
                             max_queue_size=1,
                             class_weight=class_weights,
                             callbacks=[alr_cb])
    # Elapsed time
    end = time.time()
    training_time = end - start
    print("- training time: {}s".format(training_time))
    history.history["training_time"] = training_time

    # Combine the head model with the tail model to create the full model
    model = combine_head_and_tail(model_head, model_tail)

    # Return the model and history
    return model, history



"""
Results of training
"""
import numpy as np
from sklearn.metrics import precision_recall_fscore_support, accuracy_score


class TrainingResult:
    def __init__(self,
                 model_params,
                 history,
                 y_true,
                 y_pred,
                 y_prob,
                 cls_labels,
                 training_time,
                 inference_time):

        # Model configuration
        self.model_params = model_params
        self.cls_labels = cls_labels

        # Training statistics
        self.training_time = training_time
        self.inference_time = inference_time

        # Overall accuracy
        self.accuracy = accuracy_score(y_true, y_pred)

        # Training history
        self.epochs = history.epoch
        self.loss = history.history['loss']
        self.acc = history.history.get('acc') or history.history.get('accuracy')
        if 'val_loss' in history.history.keys():
            self.val_loss = history.history['val_loss']
            self.val_acc = history.history.get('val_acc') or history.history.get('val_accuracy')
        else:
            self.val_loss = []
            self.val_acc = []

        # Accuracy metrics
        p, r, f1, s = precision_recall_fscore_support(y_true, y_pred, labels=range(len(self.cls_labels)))
        self.recall = r
        self.precision = p
        self.f1_score = f1
        self.support = s
        self.mean_precision = np.mean(self.precision)
        self.mean_recall = np.mean(self.recall)
        self.mean_f1_score = np.mean(self.f1_score)

        # Save predictions (for later analysis)
        self.y_true = y_true
        self.y_pred = y_pred
        self.y_prob = y_prob




import numpy as np
import scipy.stats as stats

class RollingBuffer:

    def __init__(self, buffer_len):
        self.__buffer = np.zeros(buffer_len)
        self.__counter = 0
        self.__buffer_len = buffer_len

    def append(self, data):
        self.__buffer = np.roll(self.__buffer, -1)
        self.__buffer[-1] = data
        self.__counter += 1
        if self.__counter > self.__buffer_len:
            self.__counter = self.__buffer_len

    def values(self):
        return self.__buffer[-self.__counter:]

    def mean(self):
        return np.sum(self.__buffer) / self.__counter

    def indices(self):
        return range(self.__counter)

    def clear(self):
        self.__counter = 0

    def length(self):
        return self.__buffer_len

    def full(self):
        return self.__counter == self.__buffer_len

    def slope_probability_less_than(self, prob):
        idxs = self.indices()
        n = len(idxs)
        if n < 3:
            return 1
        values = self.values()
        n = float(n)
        slope, intercept, r_value, p_value, std_err = stats.linregress(idxs, values)
        residuals = idxs * slope + intercept
        variance = np.sum(np.power(residuals - values, 2)) / (n - 2)
        slope_std_error = np.sqrt(variance * (12.0 / (np.power(n, 3) - n)))
        p_less_than_zero = stats.norm.cdf(prob, slope, slope_std_error)
        return p_less_than_zero
